#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Setup- Datei fuer Paralleles Cython (benutzt OpemMP).
Siehe "http://cython.readthedocs.io/en/latest/src/userguide/parallelism.html"
"""

from distutils.core import setup
from distutils.extension import Extension
from Cython.Build import cythonize

#==============================================================================
## Laut Cython Doku "http://cython.readthedocs.io/en/latest/src/userguide/parallelism.html#Compiling" muss dieser Code noch rein
ext_modules = [
      Extension(
          "cythonBeamformer",
          ["cythonBeamformer.pyx"],
          extra_compile_args=['-fopenmp'],
          extra_link_args=['-fopenmp'],
      )
 ]
setup(
     name='cythonBeamformer',
     ext_modules=cythonize(ext_modules),
 )
#==============================================================================


#==============================================================================
## Compiler optionen von Acoular
#ext_modules = [
#       Extension(
#           "cythonBeamformer",
#           ["cythonBeamformer.pyx"],
#           extra_compile_args=['-O3','-ffast-math','-msse3', \
#        '-Wno-write-strings', '-fopenmp'],
#           extra_link_args=['-lgomp'],
#       )
#  ]
#setup(
#      name='cythonBeamformer',
#      ext_modules=cythonize(ext_modules),
#  )
#==============================================================================


#==============================================================================
## Compiler optionen nach "http://nealhughes.net/parallelcomp2/"
# ext_modules=[
#     Extension("cythonBeamformer",
#               ["cythonBeamformer.pyx"],
#               libraries=["m"],
#               extra_compile_args = ["-O3", "-ffast-math", "-march=native", "-fopenmp" ],
#               extra_link_args=['-fopenmp']
#               ) 
# ]
# 
# setup( 
#   name = "cythonBeamformer",
#   cmdclass = {"build_ext": build_ext},
#   ext_modules = ext_modules
# )
#==============================================================================



from acoular import *
import numpy as np
from acoular import __file__ as bpath
from os import path
import unittest

# load some array geometry
micgeofile = path.join(path.split(bpath)[0], 'xml', 'array_64.xml')
micgeom = MicGeom(from_file=micgeofile)

num_samples = 7
block_size = 5

class TimeCacheTest(unittest.TestCase):
    global_caching_configs= ['individual','all','none','readonly','overwrite']
    def test_valid_cache_result(self):
        """
        manually create an incomplete cash file and then read it 
        """
        for conf in self.global_caching_configs:
            config.global_caching= conf
            with self.subTest(conf):
                sig = WNoiseGenerator(numsamples=num_samples)
                ps = PointSource(signal=sig, mics=micgeom)
                tc = TimeCache(source=ps)
                for i,(block_c, block_nc) in enumerate(zip(tc.result(block_size), ps.result(block_size))):
                    np.testing.assert_array_almost_equal(block_c,block_nc)
                    if i == 0: break
      
                for (block_c, block_nc) in zip(tc.result(block_size), ps.result(block_size)):
                    np.testing.assert_array_almost_equal(block_c,block_nc)


if __name__=="__main__":
    unittest.main()



import unittest
from os.path import join
import re
import numpy as np
from acoular import (
    config,
    TimeConvolve,
    WNoiseGenerator,
    PointSource,
    MicGeom,
    MaskedTimeSamples,
    tools
)
from acoular.tprocess import *

WRITE_NEW_REFERENCE_DATA = False

config.global_caching = "none"
datafile = join('..','..','examples','example_data.h5')
t1 = MaskedTimeSamples(name=datafile)
t1.start = 0 # first sample, default
t1.stop = 500 # last valid sample = 15999
invalid = list(range(4,64)) # list of invalid channels 
t1.invalid_channels = invalid # use four channels

# these are tested 
test_list = (
    "TimeInOut()",
    "TimePower()",
    "TimeAverage()",
    "TimeCumAverage()",
    "TimeReverse()",
    "Filter()",
    "FiltFiltOctave(band = 100.0, fraction = 'Third octave')",
    "FiltFiltOctave()",
    "FiltOctave(band = 100.0, fraction = 'Third octave')",
    "FiltOctave()",
    "TimeExpAverage(weight = 'F')",
    "TimeExpAverage(weight = 'S')",
    "TimeExpAverage(weight = 'I')",
    "FiltFreqWeight(weight = 'A')",
    "FiltFreqWeight(weight = 'C')",
    "FiltFreqWeight(weight = 'Z')",
    "OctaveFilterBank()"
)

def fname(s):
    """converts string to file name removing unsafe characters"""
    s1 = re.sub(r'[,.()=]', '_', s)
    return re.sub(r'[/\\:*?"<>|\' ]', '', s1)

class TprocessTest(unittest.TestCase):
    """
    A simple test case that verifies that the results of tprocess are not changing across different versions of code.
    """

    def test_timeconvolve(self):
        """compare results of timeconvolve with numpy convolve"""
        # Parameters
        NSAMPLES = 25
        N1 = WNoiseGenerator(sample_freq=1000, numsamples=NSAMPLES, seed=1)
        MGEOM = MicGeom(mpos_tot=[[1], [1], [1]])
        P1 = PointSource(signal=N1, mics=MGEOM)
        KERNEL = np.random.rand(20)
        CONV = TimeConvolve(kernel=KERNEL, source=P1)

        SIG = tools.return_result(P1, num=NSAMPLES)
        RES = tools.return_result(CONV, num=100)
        
        for i in range(P1.numchannels):
            REF = np.convolve(np.squeeze(KERNEL), np.squeeze(SIG[:,i]))
            np.testing.assert_allclose(np.squeeze(RES[:,i]), REF, rtol=1e-5, atol=1e-8)

    #@unittest.skip
    def test_tprocess_results(self):
        """compare results with reference results"""
        for s in test_list:
            b = eval(s)
            with self.subTest(s):
                name = join('reference_data',f'{fname(s)}.npy')
                b.source = t1
                # compute with block size 64 and add some extra 
                actual_data = tools.return_result(b, nmax=70, num=64)
                if WRITE_NEW_REFERENCE_DATA:
                    np.save(name,actual_data)
                ref_data = np.load(name)
                np.testing.assert_allclose(actual_data, ref_data, rtol=1e-5, atol=1e-8)

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Sep 15 11:24:36 2017

@author: tomgensch
"""

import numba as nb
import time as tm

@nb.njit
def absASquared(N):
    a = 42.0 + 87.0j
    for cnt in xrange(N):
        peter = abs(a) * abs(a)
    return peter
@nb.njit
def aTimesAConj(N):
    a = 42.0 + 87.0j
    for cnt in xrange(N):
        peter = (a * a.conjugate()).real
    return peter
@nb.njit
def realSquaredPlusImagSquared(N):
    a = 42.0 + 87.0j
    for cnt in xrange(N):
        peter = a.real * a.real + a.imag * a.imag
    return peter

N = 3000000000
t0 = tm.time()
erg1 = absASquared(N)
timeAbsASquared = tm.time() - t0

t0 = tm.time()
erg2 = aTimesAConj(N)
timeATimesAConj = tm.time() - t0

t0 = tm.time()
erg3 = realSquaredPlusImagSquared(N)
timeRealSquaredPlusImagSquared = tm.time() - t0

faktorAbsSquareVstimeRealSquaredPlusImagSquared = timeAbsASquared / timeRealSquaredPlusImagSquared
faktortimeATimesAConjVstimeRealSquaredPlusImagSquared = timeATimesAConj / timeRealSquaredPlusImagSquared

print('abs(a)**2 / (a.real**2 + a.imag**2): %s' %(faktorAbsSquareVstimeRealSquaredPlusImagSquared))
print('a*a.conj / (a.real**2 + a.imag**2): %s' %(faktortimeATimesAConjVstimeRealSquaredPlusImagSquared))



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""

"""
import time as tm

import numpy as np
from numba import guvectorize, complex128, float64

import sharedFunctions as shFncs
from beamformer import r_beamfull_os_classic, r_beamdiag_os_classic  # The benchmark (created with scipy.weave)
#from beamformer_withoutMP import r_beamfull_inverse_withoutMP  # also created with scipy.weave, but WITHOUT using multiple cores via OpenMP


#%% NUMBA - GUVECTORIZE
def csmBeamformer(csm, r0, rm, kj):
    """ Benchmark for comparison of correctness.
    """
    nFreqs = csm.shape[0]
    nGridPoints = r0.shape[0]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.complex128)
    for cntFreqs in xrange(nFreqs):
        for cntGrid in xrange(nGridPoints):
            steeringVector = np.exp(-1j * np.float32(kj[cntFreqs].imag * (rm[cntGrid, :] - r0[cntGrid])))
            # correct
            result = np.dot(steeringVector, csm[cntFreqs, :, :])
            beamformOutput[cntFreqs, cntGrid] = np.vdot(steeringVector, result)  
            
            # incorrect
#            result = np.dot(csm[cntFreqs, :, :], steeringVector)
#            beamformOutput[cntFreqs, cntGrid] = np.vdot(steeringVector, result)
    return beamformOutput.real / nMics**2


def correctResult(r0, rm, kj, eigVal, eigVec):
    nFreqs, nGridPoints, nMics = kj.shape[0], rm.shape[0], rm.shape[1]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.float64)
    for cntFreqs in xrange(nFreqs):
        for cntGrid in xrange(nGridPoints):
            steeringVector = np.exp(-1j * np.float32(kj[cntFreqs].imag * (rm[cntGrid, :] - r0[cntGrid])))
            result = 0.0
            for cntEWs in range(eigVal.shape[1]):
#                spectralCSM = np.outer(eigVec[cntFreqs, :, cntEWs], eigVec[cntFreqs, :, cntEWs].conj())
#                result += np.inner(np.inner(steeringVector.conj(), spectralCSM), steeringVector) * eigVal[cntFreqs, cntEWs]
#            beamformOutput[cntFreqs, cntGrid] = result
                result += np.vdot(eigVec[cntFreqs, :, cntEWs], steeringVector.conj()) * eigVal[cntFreqs, cntEWs] * eigVec[cntFreqs, :, cntEWs]
            beamformOutput[cntFreqs, cntGrid] = np.dot(steeringVector, result)
    return beamformOutput.real / nMics**2

def loops_NumbaGuvectorizeOverGrid(r0, rm, kj, eigVal, eigVec):
    nFreqs, nGridPoints, nMics = kj.shape[0], rm.shape[0], rm.shape[1]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.float64)
    for cntFreqs in xrange(nFreqs):
        result = np.zeros(nGridPoints, np.float64)
        loops_Core(eigVal[cntFreqs, :], eigVec[cntFreqs, :, :], r0, rm, kj[cntFreqs].imag, nMics, result)
        beamformOutput[cntFreqs, :] = result
    return beamformOutput

@guvectorize([(float64[:], complex128[:,:], float64[:], float64[:], float64[:], float64[:], float64[:])], 
              '(e),(m,e),(),(m),(),()->()', nopython=True, target='parallel')
def loops_Core(eigVal, eigVec, distGridToArrayCenter, distGridToAllMics, 
                                                             waveNumber, nMicsForNormalization, result):
    # init
    nMics = distGridToAllMics.shape[0]
    steerVec = np.zeros((nMics), np.complex128)

    # building steering vector: in order to save some operaions -> some normalization steps are applied after mat-vec-multipl.
    for cntMics in xrange(nMics):
        temp = np.float32(waveNumber[0] * (distGridToAllMics[cntMics] - distGridToArrayCenter[0]))
        steerVec[cntMics] = (np.cos(temp) - 1j * np.sin(temp))
    
    # performing matrix-vector-multiplication via spectral decomposition of the hermitian CSM-Matrix
    temp1 = 0.0
    for cntEigVal in range(len(eigVal)):
        temp2 = 0.0
        for cntMics in range(nMics):  # Algebraic multiplicity = Geometric multiplicity for any hermitian matrix
            temp2 += eigVec[cntMics, cntEigVal] * steerVec[cntMics]
        temp1 += (temp2 * temp2.conjugate() * eigVal[cntEigVal]).real
    result[0] = temp1 / nMicsForNormalization[0] ** 2


def loopsCascade_NumbaGuvectorizeOverGrid(r0, rm, kj, eigVal, eigVec):
    nFreqs, nGridPoints, nMics = kj.shape[0], rm.shape[0], rm.shape[1]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.float64)
    for cntFreqs in xrange(nFreqs):
        result = np.zeros(nGridPoints, np.float64)
        loopsCascade_Core(eigVal[cntFreqs, :], eigVec[cntFreqs, :, :], r0, rm, kj[cntFreqs].imag, nMics, result)
        beamformOutput[cntFreqs, :] = result
    return beamformOutput

@guvectorize([(float64[:], complex128[:,:], float64[:], float64[:], float64[:], float64[:], float64[:])], 
              '(e),(m,e),(),(m),(),()->()', nopython=True, target='parallel')
def loopsCascade_Core(eigVal, eigVec, distGridToArrayCenter, distGridToAllMics, 
                                                             waveNumber, nMicsForNormalization, result):
    # init
    nMics = distGridToAllMics.shape[0]
    steerVec = np.zeros((nMics), np.complex128)

    # building steering vector: in order to save some operaions -> some normalization steps are applied after mat-vec-multipl.
    for cntMics in xrange(nMics):
        temp = np.float32(waveNumber[0] * (distGridToAllMics[cntMics] - distGridToArrayCenter[0]))
        steerVec[cntMics] = (np.cos(temp) - 1j * np.sin(temp))
    
    # performing matrix-vector-multiplication via spectral decomposition of the hermitian CSM-Matrix
    temp1 = 0.0
    for cntEigVal in range(len(eigVal)):
        temp2 = np.dot(eigVec[:, cntEigVal], steerVec)
        temp1 += (temp2 * temp2.conjugate() * eigVal[cntEigVal]).real
    result[0] = temp1 / nMicsForNormalization[0] ** 2
#%%  MAIN
listOfMics = [64, 100, 250, 500, 700, 1000]
listGridPoints = [100, 5000, 10000]  # Standard value: 12000  # The number of gridpoints doesn't seeme to have to great of an influence
nTrials = 10
listOfNFreqs = [10000]

#==============================================================================
# The benchmark function 'r_beamfull_inverse' and also other implementations of 
# the beamformer create a lot of overhead, which influences the computational 
# effort of the succeding function. This is mostly the case, if concurrent
# calculations are done (multiple cores). So often the first trial of a new 
# function takes some time longer than the other trials.
#==============================================================================

#funcsToTrial = [csmBeamformer, correctResult, r_beamfull_os_classic]  # full csm
#funcsToTrial = [csmBeamformer, correctResult, r_beamdiag_os_classic]  # removed diagonal of csm
funcsToTrial = [loopsCascade_NumbaGuvectorizeOverGrid, loops_NumbaGuvectorizeOverGrid, r_beamfull_os_classic] 
removeDiagOfCSM = False

for nMics in listOfMics:
    for nGridPoints in listGridPoints:
        for nFreqs in listOfNFreqs:
            # Init
            print(10*'-' + 'New Test configuration: nMics=%s, nGridpoints=%s, nFreqs=%s' %(nMics, nGridPoints, nFreqs) + 10*'-')
            print(10*'-' + 'Creation of inputInputs' + 10*'-')
            
            # Inputs for the beamformer methods:
            # At the moment the beamformer-methods are called once per 
            # frequency (CSM is a Matrix, no 3rd-order-tensor)
            # For easier camparability we build the CSM as a 3rd-order-tensor) instead
            csm = np.random.rand(nFreqs, nMics, nMics) + 1j*np.random.rand(nFreqs, nMics, nMics)  # cross spectral matrix
            for cntFreqs in range(nFreqs):
                csm[cntFreqs, :, :] += csm[cntFreqs, :, :].T.conj()  # make CSM hermetical
            e = np.random.rand(nMics) + 1j*np.random.rand(nMics)  # has no usage
            h = np.zeros((nFreqs, nGridPoints))  # results are stored here, if function has no return value
            r0 = np.random.rand(nGridPoints)  # distance between gridpoints and middle of array
            rm = np.random.rand(nGridPoints, nMics)  # distance between gridpoints and all mics in the array
            kj = np.zeros(nFreqs) + 1j*np.random.rand(nFreqs) # complex wavenumber 
            eigVal, eigVec = np.linalg.eigh(csm)
            indLow = 5  # 5
            indHigh = 10# 10
            
            # remove diagonal of csm
            if removeDiagOfCSM:
                for cntMics in range(nMics):
                    csm[cntFreqs, cntMics, cntMics] = 0
                eigValDiag, eigVecDiag = np.linalg.eigh(csm)
                refFunc = 'r_beamdiag_os_classic'
                resultReference = correctResult(r0, rm, kj, eigValDiag[:, indLow : indHigh], eigVecDiag[:, :, indLow : indHigh])  # For relative/absolute error
            else:
                refFunc = 'r_beamfull_os_classic'
                resultReference = correctResult(r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh])  # For relative/absolute error
            
            nameOfFuncsToTrial = map(lambda x: x.__name__, funcsToTrial)
            nameOfFuncsForError = [funcName for funcName in nameOfFuncsToTrial if funcName != refFunc]
            maxRelativeDeviation = np.zeros((len(funcsToTrial), nTrials))
            maxAbsoluteDeviation = np.zeros((len(funcsToTrial), nTrials))
            timeConsumption = [[] for _ in range(len(funcsToTrial))]
            indOfBaselineFnc = nameOfFuncsToTrial.index(refFunc)
            
            # Testing
            print(10*'-' + 'Testing of functions' + 10*'-')
            cntFunc = 0
            for func in funcsToTrial:
                print(func.__name__)
                for cntTrials in xrange(nTrials):
                    h = np.zeros((nFreqs, nGridPoints))
                    if func.__name__ == refFunc:
                        t0 = tm.time()
                        func(e, h, r0, rm, kj, eigVal, eigVec, indLow, indHigh)
                        t1 = tm.time()
                        result = h / nMics**2
                    elif func.__name__ == 'correctResult' and removeDiagOfCSM:
                        t0 = tm.time()
                        output = func(r0, rm, kj, eigValDiag[:, indLow : indHigh], eigVecDiag[:, :, indLow : indHigh])
                        t1 = tm.time()
                        result = output
                    elif func.__name__ == 'csmBeamformer':
                        t0 = tm.time()
                        output = func(csm, r0, rm, kj)
                        t1 = tm.time()
                        result = output
                    else:
                        t0 = tm.time()
                        output = func(r0, rm, kj, eigVal[:, indLow : indHigh], eigVec[:, :, indLow : indHigh])
                        t1 = tm.time()
                        result = output
                    timeConsumption[cntFunc].append(t1 - t0)
                    relativeDiffBetweenNewCodeAndRef = (result - resultReference) / (result + resultReference) * 2  # error in relation to the resulting value
                    maxRelativeDeviation[cntFunc, cntTrials] = np.amax(np.amax(abs(relativeDiffBetweenNewCodeAndRef), axis=1), axis=0)  # relative error in inf-norm
                    maxAbsoluteDeviation[cntFunc, cntTrials] = np.amax(np.amax(abs(result - resultReference), axis=1), axis=0)  # absolute error in inf-norm
                cntFunc += 1
            factorTimeConsump = [np.mean(timeConsumption[cnt]) for cnt in range(0, len(funcsToTrial))] \
                                / np.mean(timeConsumption[indOfBaselineFnc])
            
            # Save the current test-config as .sav
            helpString = 'The order of the variables is: \n nameOfFuncsToTrial \n maxRelativeDeviation'\
                '\n timeConsumption [nFuncs, nTrials] \n nMics \n nGridPoints \n nFreqs '\
                '\n Factor of time consumption (in relation to the original .cpp) \n maxAbsoluteDeviation \n nThreadsGlobal'
            saveTupel = (helpString, nameOfFuncsToTrial, maxRelativeDeviation, timeConsumption, 
                         nMics, nGridPoints, nFreqs, factorTimeConsump, maxAbsoluteDeviation, 0)  
            stringParameters = 'OvernightTestcasesBeamformer_nMics%s_nGridPoints%s_nFreqs%s_nTrials%s' %(nMics, nGridPoints, nFreqs, nTrials)
            
#            stringSaveName = 'Peter'
            stringSaveName = 'Sicherung_DurchgelaufeneTests/Beamformer/cascadingSums/' + stringParameters
#            stringSaveName = 'Sicherung_DurchgelaufeneTests/Beamformer/AllImportantMethods/' + stringParameters
#            stringSaveName = 'Sicherung_DurchgelaufeneTests/Beamformer/EinflussGridpoints/AMDFX6100/' + stringParameters
#            stringSaveName = 'Sicherung_DurchgelaufeneTests/Beamformer/JitPrange/' + stringParameters
#            stringSaveName = 'Sicherung_DurchgelaufeneTests/Beamformer/Multithreading_02Threads/' + stringParameters

            shFncs.savingTimeConsumption(stringSaveName, saveTupel)  # saving as "stringSaveName.sav"
            
#            shFncs.plottingOfOvernightTestcasesBeamformer(stringSaveName + '.sav')  # plot of the current test-config

#==============================================================================
#The Following use of the numba decorators could lead to less code (as a function 
#body could be used more often) but is also slower, which is why it wasn't used
#in this comparison.
# signature = complex128[:,:](complex128[:,:,:], float64[:], float64[:,:])
# numbaOptimizedFunction= jit(signature, nopython=True)(plainPythonFunction.py_func)
#==============================================================================



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
#------------------------------------------------------------------------------
"""Simple speed benchmark tool using pyperf. Reads test* modules, and runs all
functions starting with 'test'. The results are stored per machine in the
./results subdirectory. If result already exists for a particular benchmark,
benchmark is skipped. Any command line arguments are interpreted as modules to
run. In this case, only these are considered for benchmarking. No arguments
means all test* modules are benchmarked.
"""

import glob
import types
import subprocess
import socket
import sys
from os import path, mkdir

if len(sys.argv)>1:
    testmodules = sys.argv[1:]
else:
    testmodules = glob.glob("test*.py")

machine = socket.gethostname()
if not path.exists("results"):
    mkdir("results")
if not path.exists(path.join("results", machine)):
    mkdir(path.join("results", machine))


def filt_func(f):
    return f.startswith("test") and isinstance(eval(f), types.FunctionType)


for testmodule in testmodules:
    mname = testmodule.split(".")[0]
    module = __import__(mname, globals(), locals(), [], 0)

    testfunctions = tuple(
        filter(
            lambda f: (
                f.startswith("test")
                and isinstance(module.__getattribute__(f), types.FunctionType)
            ),
            dir(module),
        )
    )
    for fname in testfunctions:
        ofnam = path.join("results", machine, f"{fname[5:]}.json")
        if path.exists(ofnam):
            print(f"{ofnam} exists, skipping")
            continue
        print(f"running {fname}:")
        subprocess.run(
            (
                "pyperf",
                "timeit",
                "-s",
                f"import {mname}",
                f"{mname}.{fname}()",
                "--append",
                ofnam,
                "--values=5",
                "--processes=2",
            )
        )



import unittest
from os.path import join
import numpy as np
from acoular import __file__ as bpath, config, WNoiseGenerator, PointSource, MicGeom

config.global_caching = "none"

# if this flag is set to True, new data will be simulated and
WRITE_NEW_REFERENCE_DATA = False
# new source results are generated for comparison during testing. Should always be False. Only set to
# true, if it is necessary to recalculate the data, due to wanted changes of the sources.

# Parameters
SFREQ = 1000
SEED = 1
NSAMPLES = 100
N1 = WNoiseGenerator(sample_freq=SFREQ, numsamples=NSAMPLES, seed=SEED)
MGEOM = MicGeom(mpos_tot=[[1],[1],[1]])


def get_source_result(Source, num=32):
    """
    returns the result for a given source

    Parameters
    ----------
    source : cls
        source class that is tested.
    num : int, optional
        number of samples to return. The default is 32.

    Returns
    -------
    array
        first block returned by the source result() function.

    """
    if Source == PointSource:
        source = Source(signal=N1, mics=MGEOM)

    return next(source.result(num)).astype(np.float32)


class SourcesTest(unittest.TestCase):
    """
    A simple test case that verifies that the results of sources are not changing across different versions of code.
    """

    sources = [PointSource]

    def test_sources_result(self):
        """compare results of sources against previous
        results from .npy file"""
        for source in self.sources:
            with self.subTest(source.__name__):
                name = join("reference_data", f"{source.__name__}.npy")
                actual_data = get_source_result(source)
                if WRITE_NEW_REFERENCE_DATA:
                    np.save(name, actual_data)
                ref_data = np.load(name)
                np.testing.assert_allclose(actual_data, ref_data, rtol=1e-5, atol=1e-8)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
Example 6 for acoular library

demonstrates different steering vectors in acoular,
and CSM diagonal removal 
with same setup as in example 1

uses measured data in file example_data.h5
calibration in file example_calib.xml
microphone geometry in array_56.xml (part of acoular)


Copyright (c) 2006-2017 The Acoular developers.
All rights reserved.
"""
from __future__ import print_function

# imports from acoular
import acoular
from acoular import L_p, Calib, MicGeom, EigSpectra, \
RectGrid, BeamformerBase, BeamformerEig, BeamformerOrth, BeamformerCleansc, \
MaskedTimeSamples, BeamformerDamas, BeamformerFunctional, PointSpreadFunction, BeamformerClean, BeamformerDamasPlus

# other imports
from os import path
from pylab import figure, subplot, imshow, show, colorbar, title, suptitle

# files
datafile = 'example_data.h5'
calibfile = 'example_calib.xml'
micgeofile = path.join( path.split(acoular.__file__)[0],'xml','array_56.xml')

#octave band of interest
cfreq = 4000

#===============================================================================
# first, we define the time samples using the MaskedTimeSamples class
# alternatively we could use the TimeSamples class that provides no masking
# of channels and samples
#===============================================================================
t1 = MaskedTimeSamples(name=datafile)
t1.start = 0 # first sample, default
t1.stop = 16000 # last valid sample = 15999
invalid = [1,7] # list of invalid channels (unwanted microphones etc.)
t1.invalid_channels = invalid 

#===============================================================================
# calibration is usually needed and can be set directly at the TimeSamples 
# object (preferred) or for frequency domain processing at the PowerSpectra 
# object (for backwards compatibility)
#===============================================================================
t1.calib = Calib(from_file=calibfile)

#===============================================================================
# the microphone geometry must have the same number of valid channels as the
# TimeSamples object has
#===============================================================================
m = MicGeom(from_file=micgeofile)
m.invalid_channels = invalid

#===============================================================================
# the grid for the beamforming map; a RectGrid3D class is also available
# (the example grid is very coarse)
#===============================================================================
g = RectGrid(x_min=-0.6, x_max=-0.0, y_min=-0.3, y_max=0.3, z=0.68,
             increment=0.05)

#===============================================================================
# for frequency domain methods, this provides the cross spectral matrix and its
# eigenvalues and eigenvectors, if only the matrix is needed then class 
# PowerSpectra can be used instead
#===============================================================================
f = EigSpectra(time_data=t1, 
               window='Hanning', overlap='50%', block_size=128, #FFT-parameters
               ind_low=7, ind_high=15, precision='complex128') #to save computational effort, only
               # frequencies with index 1-30 are used
csm = f.csm[:]
eva = f.eva[:]

f32 = EigSpectra(time_data=t1, 
               window='Hanning', overlap='50%', block_size=128, #FFT-parameters
               ind_low=7, ind_high=15, precision='complex64') #to save computational effort, only
csm32 = f32.csm[:]
eva32 = f32.eva[:]

psf32 = PointSpreadFunction(grid=g, mpos=m, c=346.04, precision='float32')
psf32Res = psf32.psf[:]

psf64 = PointSpreadFunction(grid=g, mpos=m, c=346.04, precision='float64')
psf64Res = psf64.psf[:]

bb32 = BeamformerBase(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, precision='float32')
bb32Res = bb32.synthetic(cfreq,1)

bb64 = BeamformerBase(freq_data=f, grid=g, mpos=m, r_diag=True, c=346.04, precision='float64')
bb64Res = bb64.synthetic(cfreq,1)

bf = BeamformerFunctional(freq_data=f, grid=g, mpos=m, r_diag=False, c=346.04, gamma = 60, precision='float32')
bfRes = bf.synthetic(cfreq,1)

# 32 Bit PSF precision
bd3232 = BeamformerDamas(beamformer=bb32, n_iter=100, psf_precision='float32')
bd3232Res = bd3232.synthetic(cfreq,1)
bc3232 = BeamformerClean(beamformer=bb32, psf_precision='float32')
bc3232Res = bc3232.synthetic(cfreq,1)
bdp3232 = BeamformerDamasPlus(beamformer=bb32, n_iter=100, psf_precision='float32')
bdp3232Res  = bdp3232.synthetic(cfreq,1)

#64 Bit
bd3264 = BeamformerDamas(beamformer=bb32, n_iter=100, psf_precision='float64')
bd3264Res = bd3264.synthetic(cfreq,1)
bc3264 = BeamformerClean(beamformer=bb32, psf_precision='float64')
bc3264Res = bc3264.synthetic(cfreq,1)
bdp3264 = BeamformerDamasPlus(beamformer=bb32, n_iter=100, psf_precision='float64')
bdp3264Res  = bdp3264.synthetic(cfreq,1)

# 32 Bit PSF precision
bd6432 = BeamformerDamas(beamformer=bb64, n_iter=100, psf_precision='float32')
bd6432Res = bd6432.synthetic(cfreq,1)
bc6432 = BeamformerClean(beamformer=bb64, psf_precision='float32')
bc6432Res = bc6432.synthetic(cfreq,1)
bdp6432 = BeamformerDamasPlus(beamformer=bb64, n_iter=100, psf_precision='float32')
bdp6432Res  = bdp6432.synthetic(cfreq,1)

#64 Bit
bd6464 = BeamformerDamas(beamformer=bb64, n_iter=100, psf_precision='float64')
bd6464Res = bd6464.synthetic(cfreq,1)
bc6464 = BeamformerClean(beamformer=bb64, psf_precision='float64')
bc6464Res = bc6464.synthetic(cfreq,1)
bdp6464 = BeamformerDamasPlus(beamformer=bb64, n_iter=100, psf_precision='float64')
bdp6464Res  = bdp6464.synthetic(cfreq,1)



#===============================================================================
# plot result maps for different beamformers in frequency domain
#===============================================================================
i1 = 1
for b in (bb32, bd3232, bc3232, bdp3232, bd3264, bc3264, bdp3264,
          bb64, bd6432, bc6432, bdp6432, bd6464, bc6464, bdp6464):
    subplot(2, 7, i1)
    i1 += 1
    res = b.synthetic(cfreq,1)
    mx = L_p(res.max())
    imshow(L_p(res.T), vmax=mx, vmin=mx-15, 
           interpolation='nearest', extent=g.extend())
    print(b.steer)
    colorbar()
    title(b.__class__.__name__ + b.precision,fontsize='small')


show()



#------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
#------------------------------------------------------------------------------
#
# demonstrates band pass filter characteristics and shows
# - unfiiltered power spectrum
# - three adjacent octave bands filtered spectra
# - sum of all octave band spectra
# 
# ----------------------------------------------------------------

from acoular import (
    config,
    WNoiseGenerator,
    MicGeom,
    PointSource,
    FiltFiltOctave,
    PowerSpectra,
    tools
)
import numpy as np
import matplotlib.pyplot as pl
config.global_caching = 'none'
numsamples = 51200*10
n1 = WNoiseGenerator(sample_freq=51200, numsamples=numsamples, seed=1)
m1 = MicGeom(mpos_tot=[[1], [1], [1]])
p1 = PointSource(signal=n1, mics=m1)
f1 = FiltFiltOctave(source = p1, band = 10, fraction = 'Octave')
hs = 0.0
res0 = 0.0
for i in np.arange(2.1,3.0,0.3):
    #f1.order = 2
    f1.band = 10**i
    res = (tools.return_result(f1)**2).sum()
    res0 += res
    ps = PowerSpectra(time_data=f1,block_size=8192)
    print(res/numsamples,ps.csm[:,0,0].real.sum())
    x = ps.fftfreq()
    y = 10*np.log10(ps.csm[:,0,0].real)
    pl.semilogx(x,y)
    hs = hs + ps.csm[:,0,0].real
y = 10 * np.log10(abs(hs))
pl.semilogx(x,y)
ps = PowerSpectra(time_data=p1,block_size=8192)
x = ps.fftfreq()
y = 10*np.log10(ps.csm[:,0,0].real)
pl.semilogx(x,y)
pl.ylim(-45,-35)
pl.xlim(70,1400)
pl.show()



import unittest
import acoular as ac
from copy import deepcopy
class GridTest(unittest.TestCase):
   
    @staticmethod
    def get_sector_classes():
        # for later testing condition: sector only includes (0,0,1) point 
        sectors = [
            ac.CircSector(x=0,y=0,r=0.2), 
            ac.RectSector(x_min=-0.2,x_max=0.2,y_min=-0.2,y_max=0.2),
            ac.RectSector3D(x_min=-0.2,x_max=0.2,y_min=-0.2,y_max=0.2,z_min=1,z_max=1),
            ac.PolySector(edges=[0.2,0.2,-0.2,0.2,-0.2,-0.2,0.2,-0.2]),
            ac.ConvexSector(edges=[0.2,0.2,-0.2,0.2,-0.2,-0.2,0.2,-0.2]),            
            ]
        multi_sector = ac.MultiSector(sectors=deepcopy(sectors))
        return sectors + [multi_sector]

    @staticmethod
    def get_emtpy_sector_classes():
        # for later testing condition: sector should not cover any grid point
        off = 10
        sectors = [
            ac.CircSector(x=off,y=off,r=0.2, include_border=False, default_nearest=False),
            ac.RectSector(
                x_min=-0.2+off,x_max=0.2+off,y_min=-0.2+off,y_max=0.2+off, 
                include_border=False, default_nearest=False), 
            ac.RectSector3D(
                x_min=-0.2+off,x_max=0.2+off,y_min=-0.2+off,y_max=0.2+off,z_min=1+off,z_max=1+off, 
                include_border=False, default_nearest=False),               
            ac.PolySector(
                edges=[0.2+off,0.2+off,-0.2+off,0.2+off,-0.2+off,-0.2+off,0.2+off,-0.2+off], 
                include_border=False, default_nearest=False),
            ac.ConvexSector(
                edges=[0.2+off,0.2+off,-0.2+off,0.2+off,-0.2+off,-0.2+off,0.2+off,-0.2+off], 
                include_border=False, default_nearest=False)]
        multi_sector = ac.MultiSector(
            sectors=deepcopy(sectors))
        return sectors + [multi_sector]

    @staticmethod
    def get_rectgrid():
        return ac.RectGrid(x_min=-1,x_max=1,y_min=-1,y_max=1,z=1,increment=1)
        
    @staticmethod
    def get_rectgrid3D():
        return ac.RectGrid3D(x_min=-1,x_max=1,y_min=-1,y_max=1,z_min=1, z_max=1,increment=1)

    @staticmethod
    def get_linegrid():
        return ac.LineGrid(loc=(-1,0,1), length=2, numpoints=3)

    @staticmethod
    def get_importgrid():
        return ac.ImportGrid(gpos_file=GridTest.get_rectgrid().gpos)

    @staticmethod
    def get_mergegrid():
        return ac.MergeGrid(grids=[GridTest.get_rectgrid(), GridTest.get_linegrid()])
    
    @staticmethod
    def get_all_grids():
        for grid in [GridTest.get_rectgrid, GridTest.get_rectgrid3D,
            GridTest.get_linegrid, GridTest.get_importgrid, GridTest.get_mergegrid]:
            yield grid()

    def test_size(self):
        for grid in self.get_all_grids():
            with self.subTest(grid.__class__.__name__):
                self.assertEqual(grid.size, grid.gpos.shape[-1])

    def test_existing_subdomain(self):
        for grid in self.get_all_grids():
            for sector in self.get_sector_classes():
                with self.subTest(f"Grid: {grid.__class__.__name__} Sector:{sector.__class__.__name__}"):
                    indices = grid.subdomain(sector)
                    self.assertEqual(indices[0].shape[0], 1)

    def test_empty_subdomain(self):
        for grid in self.get_all_grids():
            for sector in self.get_emtpy_sector_classes():
                with self.subTest(f"Grid: {grid.__class__.__name__} Sector:{sector.__class__.__name__}"):
                    indices = grid.subdomain(sector)
                    self.assertEqual(indices[0].shape[0], 0)
                    if hasattr(sector, 'default_nearest'):
                        sector.default_nearest = True
                        indices = grid.subdomain(sector)
                        self.assertEqual(indices[0].shape[0], 1)
                   

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Wed May 24 11:09:29 2017

@author: tomgensch
"""
from cPickle import dump, load
import matplotlib.pylab as plt
import numpy as np
import glob


def savingTimeConsumption(fileName, saveTuple):
    """ Saves all data in 'saveTuple' into 'fileName.sav'. 
    The First entry of the tuple has to be a String, explaining the structure of 
    the tuple."""
    fi = open(fileName + '.sav', 'w')
    dump(saveTuple, fi, -1)
    fi.close()
    return 0


def readingInSAVES(fileName):
    """ Reads in the Data saved with 'savingTimeConsumption'. """
    fi = open(fileName, 'r')
    data = load(fi)
    fi.close()
    helpText = data[0]
    returnData = data[1:]
    return helpText, returnData


def plottingTimeConsumptions(titleString, trialedFuncs, timesToPlot):
    """ titleString...String to be displayed in Title
    trialedFuncs...list of the strings of the trialed functions
    timesToPlot...dim [numberTrials, numberFunctions]
    """
#    plt.figure()
    for cnt in range(len(trialedFuncs)):
        if 'vectorized' in trialedFuncs[cnt]:
            lineStyle = '--'
        elif 'faverage' in trialedFuncs[cnt]:
            lineStyle = '--'
        else:
            lineStyle = '-'
        plt.semilogy(timesToPlot[cnt], label=trialedFuncs[cnt], linestyle=lineStyle, marker='o')
    plt.xticks(range(len(timesToPlot[1])))
    plt.xlabel('trials [1]')
    plt.ylabel('Time per Trial [s]')
    plt.grid(which='major')
    plt.grid(which='minor', linestyle='--')
    plt.title(titleString)
    yMin, yMax = plt.ylim()
    newYMin = 10 ** np.floor(np.log10(yMin))
    plt.ylim(newYMin, yMax)
    plt.legend(bbox_to_anchor=(1.01, 1), loc=2, borderaxespad=0.)
    plt.show()


def plottingOfOvernightTestcasesBeamformer(fileName):
    helpText, daten = readingInSAVES(fileName)
    
    stringForLegend = []
    hFig = plt.figure()
    hAxesRelErr = hFig.add_subplot(3,2,5)
    hAxesRelErr.set_ylabel('relative Error - infNorm')
    hAxesRelErr.set_xlabel('trials [1]')
    
    hAxesAbsErr = hFig.add_subplot(3,2,6)
    hAxesAbsErr.set_ylabel('absolute Error - infNorm')
    hAxesAbsErr.set_xlabel('trials [1]')
    
    numberMethods = len(daten[0])
    numberTrials = len(daten[1][0, :])
    trials = np.arange(numberTrials)
    
    # For barplot
    withPerTrial = 0.75
    widthBar = withPerTrial / numberMethods
    offsetOfXAxes = withPerTrial / 2 - widthBar

    # plotting error
    for cnt in xrange(numberMethods):
        stringForLegend.append(daten[0][cnt] + ' | Time=%1.2f' %(daten[6][cnt]))  # For time consumption
        hAxesRelErr.bar(trials + cnt * widthBar - offsetOfXAxes, daten[1][cnt, :], widthBar, label=daten[0][cnt])  # relative error
        hAxesAbsErr.bar(trials + cnt * widthBar - offsetOfXAxes, daten[7][cnt, :], widthBar, label=daten[0][cnt])  # absolute error
    hAxesAbsErr.legend()
    hAxesAbsErr.set_yscale('log')
    hAxesAbsErr.set_xticks(trials)
    
    hAxesRelErr.set_yscale('log')
    hAxesRelErr.set_xticks(trials)
    
    # plotting time consumption
    titelString = 'Performance Comparison, nMics = %s, nGridPoints = %s, nFreqs = %s.'\
                '\n With time consumption factor in relation to the \noriginal r_beamfull_inverse'\
                ' in the legend'\
                '\n If a method works with manually spawn threads: nThreads = %s.'%(daten[3], daten[4], daten[5], daten[8])
    plt.subplot(3,3,(1,5))
    plottingTimeConsumptions(titelString, stringForLegend, daten[2])
    hFig.canvas.set_window_title(fileName)


def plottingTimeConsumptionOverSpecificOrdinate(dirName, ordinate='nMics'):
    listOfFiles = glob.glob(dirName + '/*.sav')
    helpText, daten = readingInSAVES(listOfFiles[0])
    arrayOrdinate = np.zeros(len(listOfFiles))
    arrayTimeConsump = np.zeros((len(listOfFiles), len(daten[6])))
    cnt = 0
    for currentfile in listOfFiles:
        helpText, daten = readingInSAVES(currentfile)
        if ordinate == 'nMics':
            arrayOrdinate[cnt] = daten[3]
        arrayTimeConsump[cnt, :] = daten[6]
        cnt += 1
    indSorted = np.argsort(arrayOrdinate)
    plt.semilogy(arrayOrdinate[indSorted], arrayTimeConsump[indSorted, :], marker='o')#, label=trialedFuncs[cnt], linestyle=lineStyle, marker='o')
    plt.legend(daten[0])
    plt.grid(which='major')
    plt.grid(which='minor', linestyle='--')
    plt.xlabel(ordinate)
    plt.ylabel('Mean of Time per Trial [s] (normalized to faverage)')
    plt.title('Mean of TimeConsumption over ' + ordinate + '\n asd')
    plt.xticks(arrayOrdinate)
        

def plottingOfOvernightTestcasesOnFAVERAGE(fileName):
    helpText, daten = readingInSAVES(fileName)
    titleString = 'NUMBA - using "faverage"\n' \
        'nAverages=%s,  nFreqbins=%s,  nMics=%s,  nTest=%s' % (daten[2], daten[3], daten[4], daten[5])
    plottingTimeConsumptions(titleString, daten[0], daten[1])


def plotAllAvailableTestCases(dirName):
    listOfFiles = glob.glob(dirName + '/*.sav')
    for currentfile in listOfFiles:
        try:
            plottingOfOvernightTestcasesBeamformer(currentfile)
        except:
            print('Could not plot Testcase:' + currentfile)

def saveAllCurrentlyOpenedFigures():
    for cntFig in plt.get_fignums():
        saveNameForPNGHelp = plt.figure(cntFig).canvas.get_window_title()
        saveNameForPNG = saveNameForPNGHelp.replace('.sav', '.png')
        plt.savefig(saveNameForPNG)

#plottingTimeConsumptionOverSpecificOrdinate('Sicherung_DurchgelaufeneTests/faverage/InfluenceOfMics/')
#plottingOfOvernightTestcasesBeamformer('Sicherung_DurchgelaufeneTests/.sav')
#plotAllAvailableTestCases('Sicherung_DurchgelaufeneTests/damasSolver/')
#plottingOfOvernightTestcasesBeamformer('Peter.sav')
#saveAllCurrentlyOpenedFigures()



from acoular import UniformFlowEnvironment,SlotJet,OpenJet,RotatingFlow,MicGeom, SteeringVector,\
    BeamformerOrth, RectGrid, MaskedTimeSamples, Sector,\
        UncorrelatedNoiseSource, SourceMixer, SamplesGenerator, BeamformerTimeTraj, BeamformerTimeSqTraj,\
            BeamformerCleantTraj, BeamformerCleantSqTraj, IntegratorSectorTime, MaskedTimeInOut, ChannelMixer,\
                SpatialInterpolator, SpatialInterpolatorRotation, SpatialInterpolatorConstantRotation, Mixer,\
                    WriteWAV, MergeGrid, FiltWNoiseGenerator, SphericalHarmonicSource, PointSource
from numpy import array
from unittest import TestCase

# a dictionary containing all classes that should change their digest on
# changes of the following trait types:
#   * List
#   * CArray 
UNEQUAL_DIGEST_TEST_DICT = {
#    "MicGeom.mpos_tot item assignment" : (MicGeom(mpos_tot=[[1.,2.,3.]]), "obj.mpos_tot[:] = 0."),
    "MicGeom.mpos_tot new array assignment" : (MicGeom(mpos_tot=[[1.,2.,3.]]), "obj.mpos_tot = array([0.])"),
#    "MicGeom.invalid_channels item assignment" : (MicGeom(mpos_tot=[[1.,2.,3.]],invalid_channels=[1]), "obj.invalid_channels[0] = 0"),
    "MicGeom.invalid_channels new list assignment" : (MicGeom(mpos_tot=[[1.,2.,3.]],invalid_channels=[1]), "obj.invalid_channels = [0]"),
    # environments.py
#    "UniformFlowEnvironment.fdv item assignment": (UniformFlowEnvironment(), "obj.fdv[0] = 0."),
    "UniformFlowEnvironment.fdv array assignment": (UniformFlowEnvironment(), "obj.fdv = array((0., 0., 0.))"),
#    "SlotJet.origin item assignment": (SlotJet(), "obj.origin[0] = 1."),
    "SlotJet.origin array assignment": (SlotJet(), "obj.origin = array((1., 0., 0.))"),
#    "SlotJet.flow item assignment": (SlotJet(), "obj.flow[0] = 0."),
    "SlotJet.flow array assignment": (SlotJet(), "obj.flow = array((0., 0., 0.))"),    
#    "SlotJet.plane item assignment": (SlotJet(), "obj.plane[0] = 1."),
    "SlotJet.plane array assignment": (SlotJet(), "obj.plane = array((1., 0., 0.))"),
#    "OpenJet.origin item assignment": (OpenJet(), "obj.origin[0] = 1."),
    "OpenJet.origin array assignment": (OpenJet(), "obj.origin = array((1., 0., 0.))"),   
#    "RotatingFlow.origin item assignment": (RotatingFlow(), "obj.origin[0] = 1."),
    "RotatingFlow.origin array assignment": (RotatingFlow(), "obj.origin = array((1., 0., 0.))"),  
    #fbeamform.py
#    "SteeringVector.ref item assignment": (SteeringVector(), "obj.ref[0] = 1."),  
    "SteeringVector.ref array assignment": (SteeringVector(), "obj.ref = array((1., 1., 1.))"),  
#    "BeamformerOrth.eva_list item assignment": (BeamformerOrth(eva_list=array((0, 1))), "obj.eva_list[0] = 2"),  
    "BeamformerOrth.eva_list array assignment": (BeamformerOrth(eva_list=array((0, 1))), "obj.eva_list = array((2))"),  
    #grids.py
    "MergeGrid.grids item assignment": (MergeGrid(grids=[RectGrid()]), "obj.grids[0] = RectGrid()"),  
    "MergeGrid.grids list assignment": (MergeGrid(), "obj.grids = [RectGrid()]"),
    # signals.py
#    "FiltWNoiseGenerator.ar item assignment": (FiltWNoiseGenerator(ar=[1.,2.,3.]), "obj.ar[0] = 0."),
    "FiltWNoiseGenerator.ar array assignment": (FiltWNoiseGenerator(ar=[1.,2.,3.]), "obj.ar = array((0., 0., 0.))"),   
#    "FiltWNoiseGenerator.ma item assignment": (FiltWNoiseGenerator(ma=[1.,2.,3.]), "obj.ma[0] = 0."),
    "FiltWNoiseGenerator.mar array assignment": (FiltWNoiseGenerator(ma=[1.,2.,3.]), "obj.ma = array((0., 0., 0.))"), 
    #sources.py 
    "MaskedTimeSamples.invalid_channels item assignment": (MaskedTimeSamples(invalid_channels=[1]), "obj.invalid_channels[0] = 0"),  
    "MaskedTimeSamples.invalid_channels list assignment": (MaskedTimeSamples(), "obj.invalid_channels = [0]"),
#    "SphericalHarmonicSource.alpha item assignment": (SphericalHarmonicSource(alpha=array((0, 1))), "obj.alpha[0] = 1."),  
    "SphericalHarmonicSource.alpha array assignment": (SphericalHarmonicSource(alpha=array((0, 1))), "obj.alpha = array((1., 1., 1.))"),  
#    "UncorrelatedNoiseSource.seed item assignment": (UncorrelatedNoiseSource(seed=array((1, 2))), "obj.seed[0] = 3"),  
    "UncorrelatedNoiseSource.seed array assignment": (UncorrelatedNoiseSource(seed=array((1, 2))), "obj.seed = array((3,4))"),  
    "SourceMixer.sources item assignment": (SourceMixer(sources=[SamplesGenerator()]), "obj.sources[0] = PointSource()"),  
    "SourceMixer.sources list assignment": (SourceMixer(sources=[SamplesGenerator()]), "obj.sources = [PointSource()]"),  
    # tbeamform.py
#    "BeamformerTimeTraj.rvec item assignment": (BeamformerTimeTraj(), "obj.rvec[0] = 1."),
    "BeamformerTimeTraj.rvec array assignment": (BeamformerTimeTraj(), "obj.rvec = array((1., 0., 0.))"),
#    "BeamformerTimeSqTraj.rvec item assignment": (BeamformerTimeSqTraj(), "obj.rvec[0] = 1."),
    "BeamformerTimeSqTraj.rvec array assignment": (BeamformerTimeSqTraj(), "obj.rvec = array((1., 0., 0.))"),
#    "BeamformerCleantTraj.rvec item assignment": (BeamformerCleantTraj(), "obj.rvec[0] = 1."),
    "BeamformerCleantTraj.rvec array assignment": (BeamformerCleantTraj(), "obj.rvec = array((1., 0., 0.))"),
#    "BeamformerCleantSqTraj.rvec item assignment": (BeamformerCleantSqTraj(), "obj.rvec[0] = 1."),
    "BeamformerCleantSqTraj.rvec array assignment": (BeamformerCleantSqTraj(), "obj.rvec = array((1., 0., 0.))"),
    "IntegratorSectorTime.sectors item assignment": (IntegratorSectorTime(sectors=[Sector()]), "obj.sectors[0] = Sector()"),  
    "IntegratorSectorTime.sectors list assignment": (IntegratorSectorTime(sectors=[Sector()]), "obj.sectors = [Sector()]"),  
    # tprocess.py
    "MaskedTimeInOut.invalid_channels item assignment": (MaskedTimeInOut(invalid_channels=[1]), "obj.invalid_channels[0] = 0"),  
    "MaskedTimeInOut.invalid_channels list assignment": (MaskedTimeInOut(), "obj.invalid_channels = [0]"),
#    "ChannelMixer.weights item assignment" : (ChannelMixer(weights=[[1.,2.,3.]]), "obj.weights[0] = 0."),
    "ChannelMixer.weights new array assignment" : (ChannelMixer(weights=[[1.,2.,3.]]), "obj.weights = array([0.])"),
#    "SpatialInterpolator.Q item assignment": (SpatialInterpolator(), "obj.Q[0] = 0."),
    "SpatialInterpolator.Q array assignment": (SpatialInterpolator(), "obj.Q = array([[0.,0.,0.],[0.,0.,0.],[0.,0.,0.]])"),
#    "SpatialInterpolatorRotation.Q item assignment": (SpatialInterpolatorRotation(), "obj.Q[0] = 0."),
    "SpatialInterpolatorRotation.Q array assignment": (SpatialInterpolatorRotation(), "obj.Q = array([[0.,0.,0.],[0.,0.,0.],[0.,0.,0.]])"),
#    "SpatialInterpolatorConstantRotation.Q item assignment": (SpatialInterpolatorConstantRotation(), "obj.Q[0] = 0."),
    "SpatialInterpolatorConstantRotation.Q array assignment": (SpatialInterpolatorConstantRotation(), "obj.Q = array([[0.,0.,0.],[0.,0.,0.],[0.,0.,0.]])"),
    "Mixer.sources item assignment": (Mixer(source=SamplesGenerator(), sources=[SamplesGenerator()]), "obj.sources[0] = SamplesGenerator(numchannels=1)"),  
    "Mixer.sources list assignment": (Mixer(sources=[SamplesGenerator()]), "obj.sources = [SamplesGenerator(), SamplesGenerator()]"),  
    "WriteWAV.channels item assignment": (WriteWAV(channels=[1]), "obj.channels[0] = 0"),  
    "WriteWAV.channels list assignment": (WriteWAV(), "obj.channels = [0]"),
    }


class Test_DigestChange(TestCase):
    """Test that ensures that digest of Acoular classes changes correctly on 
    changes of CArray and List attributes.
    """

    def get_digests(self,obj,statement):
        """A function that collects the digest of the obj before (d1) and
        after (d2) executing the statement that should yield to a change
        of the object digest.

        Parameters
        ----------
        obj : instance
            class instance that has an attribute `digest`
        statement : "str"
            a string that can be executed 

        Returns
        -------
        str, str
            digest before and after statement execution
        """
        d1 = obj.digest
        exec(statement)
        d2 = obj.digest
        return d1,d2

    def test_digest_changes_on_assignment(self):
        """ test that object digest change on statement execution """
        for test_label,(obj,statement) in UNEQUAL_DIGEST_TEST_DICT.items():
            # use subtest to get individual test result.
            # tests will not stop if one subtest fails!
            with self.subTest(test_label):
                digest_before_statement,digest_after_statement = self.get_digests(obj, statement)
                self.assertNotEqual(digest_before_statement,digest_after_statement)        



if __name__ == '__main__':
    import unittest 
    unittest.main()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 24 09:58:04 2019

@author: newuser
"""
#standard testing suite from python
import unittest

#acoular imports
import acoular

from acoular import L_p, Calib, MicGeom, PowerSpectra, Environment, \
RectGrid, BeamformerBase, BeamformerEig, BeamformerOrth, BeamformerCleansc, \
MaskedTimeSamples, FiltFiltOctave, BeamformerTimeSq, TimeAverage, \
TimeCache, BeamformerTime, TimePower, BeamformerCMF, \
BeamformerCapon, BeamformerMusic, BeamformerDamas, BeamformerClean, \
BeamformerFunctional, BeamformerDamasPlus, BeamformerGIB, SteeringVector

from numpy import zeros, empty
from os import path
import tables

#load numerical values from Examples
h5file_num = tables.open_file('reference_data/Example1_numerical_values_testsum.h5', 'r')

mpos_num = h5file_num.get_node('/mpos_values').read()
grid_pos_num = h5file_num.get_node('/grid_pos_values').read()
transfer_num = h5file_num.get_node('/transfer_values').read()
csm_num = h5file_num.get_node('/csm_values').read()
eve_num = h5file_num.get_node('/eva_values').read()
eva_num = h5file_num.get_node('/eve_values').read()

d={}
for b in ('bb', 'bc', 'be', 'bm', 'bl', 'bo', 'bs', 'bd', 'bcmf', 'bf', 'bdp', 'bgib'):
    d[b+'num'] = h5file_num.get_node('/'+str(b)+'_values').read()


#load exampledata
datafile = '../../examples/example_data.h5'
calibfile = '../../examples/example_calib.xml'
micgeofile = path.join( path.split(acoular.__file__)[0],'xml','array_56.xml')

#calc all values from example 1
cfreq = 4000
t1 = MaskedTimeSamples(name= datafile)
t1.start = 0 # first sample, default
t1.stop = 16000 # last valid sample = 15999
invalid = [1,7] # list of invalid channels (unwanted microphones etc.)
t1.invalid_channels = invalid 
t1.calib = Calib(from_file=calibfile)
m = MicGeom(from_file=micgeofile)
m.invalid_channels = invalid

g = RectGrid(x_min=-0.6, x_max=-0.0, y_min=-0.3, y_max=0.3, z=0.68,
             increment=0.05 )

env=Environment(c=346.04)

st = SteeringVector(grid=g, mics=m, env=env)

f = PowerSpectra(time_data=t1, 
               window='Hanning', overlap='50%', block_size=128, #FFT-parameters
               cached = False )  #cached = False


bb = BeamformerBase(freq_data=f, steer=st, r_diag=True, cached = False)
bc = BeamformerCapon(freq_data=f, steer=st, cached=False)
be = BeamformerEig(freq_data=f, steer=st, r_diag=True, n=54, cached = False)
bm = BeamformerMusic(freq_data=f, steer=st, n=6, cached = False)
bd = BeamformerDamas(beamformer=bb, n_iter=100, cached = False)
bdp = BeamformerDamasPlus(beamformer=bb, n_iter=100, cached = False)
bo = BeamformerOrth(beamformer=be, eva_list=list(range(38,54)), cached = False)
bs = BeamformerCleansc(freq_data=f, steer=st, r_diag=True, cached = False)
bcmf = BeamformerCMF(freq_data=f, steer=st, method='LassoLarsBIC', cached = False)
bl = BeamformerClean(beamformer=bb, n_iter=100, cached = False)
bf = BeamformerFunctional(freq_data=f, steer=st, r_diag=False, gamma=4, cached = False)
bgib = BeamformerGIB(freq_data=f, steer=st, method= 'LassoLars', n=10, cached = False)

class acoular_test(unittest.TestCase):  
    
    #test if microfon positions are correct
    def test_mic_positions(self):
        self.assertAlmostEqual(m.mpos.sum()/mpos_num.sum(),1,3) 
    
    #test if grid points are correct
    def test_grid_positions(self):
        self.assertAlmostEqual(g.gpos.sum()/grid_pos_num.sum(),1,3) 
        
    #test steering vector calculation
    def test_steering(self):
        ind=0
        checktrans=0
        for freq in f.fftfreq()[1:-1]:
            checktrans=checktrans+st.transfer(freq)
            ind+=1
        self.assertAlmostEqual(checktrans.sum()/transfer_num.sum(),1,3)             
        
    #test if csm values are correct
    @unittest.skip
    def test_csm_calculation(self):
        self.assertAlmostEqual(f.csm[:].sum()/csm_num.sum(),1,3)    
        
    #test eve/eva
    @unittest.skip    
    def test_eigenvalue_calculation(self):
        self.assertAlmostEqual(f.eva[:].sum()/eve_num.sum(),1,3)
        print(f.eve[:].sum(),eva_num)      
        self.assertAlmostEqual(f.eve[:].sum()/eva_num.sum(),1,3) 
     
    #test beamformer results
    @unittest.skip  
    def test_beamformer_calculation(self):
        for beam,bfname in zip((bb, bc, be, bm, bl, bo, bs, bd, bcmf, bf, bdp, bgib),('bb', 'bc', 'be', 'bm', 'bl', 'bo', 'bs', 'bd', 'bcmf', 'bf', 'bdp', 'bgib')):   
            with self.subTest(bfname):
                self.assertAlmostEqual(beam.synthetic(cfreq,1).sum()/d[bfname+'num'].sum(),1,3)      
            
if "__main__" == __name__:
    unittest.main() #exit=False


                            
        


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Comparison of different optimization approaches to the 'r_beamfull_inverse' method.
Compared are: Numpy (matrix-vector-calculations), Numba and Cython. 

1. Background: 
Currently (python=2) performance-critical acoular methods (e.g.
faverage, all freqency-domain-beamformers, ...) are optimized via Scipy.weave
which translates code to c++, including compiling. Those executables can then
be imported in python.
Scipy.weave isn't supported anymore in python=3. Furthermore the executables of
already with Scipy.weave (python=2) compiled code cannot be imported in python=3.

2. Structure of comparison:
The benchmark in both errors and time consumption is always the Scipy.weave build, 
OpenMP optimized, c++ compiled code. Especially the relative and absolute inf-norm 
errors in the plots refer to the outputs of this function.

3. Remarks to the code:
In various codes below there are repeating patterns which may need some explanation:
a. There is a cast from 64-bit double precision to 32-bit precision of 'temp3',
    which is the argument to the exp() when calculating the steering vectors 
    -> This down-cast shortens the series expansion of exp() drastically which 
    leads to faster calculations while having acceptable errors. In fact if 
    there is no down-cast, the relative error between otherwise identical 
    methods is about 10^-8.
b. The exp() (when calculating the steering vector, see a.) is mostly replaced
    by a direct calculating of 'cos() - 1j*sin()', which can be done because
    the input 'temp3' of exp(temp3) is pure imaginary. Because of this the 
    calculation of exp(0)=1 in 'exp(0 - 1j*a) = exp(0) * (cos(a) - 1j*sin(a))'
    can be spared. This leads to further speed improvements.

4. Remark on the use of Cython:
See file 'cythonBeamformer.pyx' for remarks on Cython. It showed that at the 
moment Cython doesn't work to well for the beamformer case.

5. Others:


Versions used in this script:
numba=0.34.0
python=2.7.13



  # multiplizieren mit nMics erfolgt ausserhalb, sonst hier



"""
import time as tm
import threading
import gc

import numpy as np
from numba import jit, guvectorize, complex128, complex64, float64, float32, void, uint64, njit, prange

import sharedFunctions as shFncs
from cythonBeamformer import beamformerCython, beamformerCythonNOTparallel  # created with cython
from beamformer import r_beamfull_inverse  # The benchmark (created with scipy.weave)
from beamformer_withoutMP import r_beamfull_inverse_OhneMP  # also created with scipy.weave, but WITHOUT using multiple cores via OpenMP

#%% Formulate the Beamformer as VECTOR * MATRIX * VECTOR product
def vectorized(csm, e, h, r0, rm, kj):
    """ Uses Numpys fast array operations, distributed via the mkl-package.
    Those oparations are already optimized and use all available physical cores.
    """
    nFreqs = csm.shape[0]
    nGridPoints = r0.shape[0]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.complex128)
    for cntFreqs in xrange(nFreqs):
        for cntGrid in xrange(nGridPoints):
            steeringVector = rm[cntGrid, :] / r0[cntGrid] \
                             * np.exp(-1j * kj[cntFreqs].imag * (rm[cntGrid, :] - r0[cntGrid]))
            beamformOutput[cntFreqs, cntGrid] = np.inner(np.inner(steeringVector.T.conj(), csm[cntFreqs, :, :]), steeringVector)
    return beamformOutput.real

def vectorizedOptimized(csm, e, h, r0, rm, kj):
    """ Same as 'vectorized' but including both 3.a. & 3.b. of the documentation
    string at the beginning of this file. In opposite to the numba-optimized
    methods below, the use of 'cos() - 1j*sin()' instead of 'exp()' (see 3.a.)
    doesn't seem to have any speed improvement here.    
    """
    nFreqs = csm.shape[0]
    nGridPoints = r0.shape[0]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.complex128)
    for cntFreqs in xrange(nFreqs):
        kjj = kj[cntFreqs].imag
        for cntGrid in xrange(nGridPoints):
            temp3 = np.float32(kjj * (rm[cntGrid, :] - r0[cntGrid]))
            steeringVector = rm[cntGrid, :] / r0[cntGrid] * (np.cos(temp3) - 1j * np.sin(temp3))
            beamformOutput[cntFreqs, cntGrid] = np.vdot(steeringVector, np.dot(steeringVector, csm[cntFreqs, :, :]))
    return beamformOutput.real

@jit
def vectorized_NumbaJitOnly(csm, e, h, r0, rm, kj):
    """ Identical code to vectorized. Just decorated with the most basic 
    jit-optimization routine. If jit is able to translate all variables into
    primitive datatypes (NOT the native python objects) it will do that. If not,
    jit will fall back into 'Object mode' (native python objects) which will 
    mostly be much slower.
    """
    nFreqs = csm.shape[0]
    nGridPoints = r0.shape[0]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.complex128)
    for cntFreqs in xrange(nFreqs):
        for cntGrid in xrange(nGridPoints):
            steeringVector = rm[cntGrid, :] / r0[cntGrid] \
                             * np.exp(-1j * kj[cntFreqs].imag * (rm[cntGrid, :] - r0[cntGrid]))
            beamformOutput[cntFreqs, cntGrid] = np.inner(np.inner(steeringVector.T.conj(), csm[cntFreqs, :, :]), steeringVector)
    return beamformOutput.real

@jit(nopython=True)  # same as directly calling @njit
def vectorized_NumbaJit_nopythonTrue(csm, e, h, r0, rm, kj):
    """ In addition to 'vectorized_NumbaJitOnly' the nopython=True (or simply 
    @njit for numby>=0.34.) makes shure that if jit cannot translate the code 
    into primitive datatypes, it will NOT fall back into object mode but 
    instead returns an error.
    """
    nFreqs = csm.shape[0]
    nGridPoints = r0.shape[0]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.complex128)
    for cntFreqs in xrange(nFreqs):
        for cntGrid in xrange(nGridPoints):
            steeringVector = rm[cntGrid, :] / r0[cntGrid] * np.exp(-1j * kj[cntFreqs].imag * (rm[cntGrid, :] - r0[cntGrid]))
            
#==============================================================================
#             #Not all numpy functions are supported in jit. It took some 
#             #tries to figure out how to implement the easy np.inner used in 
#             #'vectorized' into jit-supported functions.
#             

#             peer = np.inner(np.inner(steeringVector.T.conjugate(), csm[cntFreqs, :, :]), steeringVector)
#             peer2 = np.vdot(steeringVector, np.dot(steeringVector, csm[cntFreqs, :, :]))  # Der scheint zu klappen 
#             peer3 = np.vdot(steeringVector, np.dot(csm[cntFreqs, :, :], steeringVector))
#             
#             versuch1 = np.inner(steeringVector, csm[cntFreqs, :, :])
#             versuch12 = np.inner(steeringVector.conjugate(), csm[cntFreqs, :, :])
#             versuch2 = np.dot(steeringVector, csm[cntFreqs, :, :])  # complex konj von versuch 12
#             versuch3 = np.dot(csm[cntFreqs, :, :], steeringVector)  # gleiche wie versuch1
#             versuch4 = np.dot(steeringVector.conjugate(), csm[cntFreqs, :, :])  # is das kompl conjugierte zu versuch1, versuch3
#             versuch5 = np.dot(csm[cntFreqs, :, :], steeringVector.conjugate())  # is das gleiche wie versuch12
#             ##--> Anscheinend ist die Syntax fuer x^H * A = dot(A, x.conj)
            beamformOutput[cntFreqs, cntGrid] = np.vdot(steeringVector, np.dot(steeringVector, csm[cntFreqs, :, :]))  # This works
#==============================================================================
    return beamformOutput.real

@njit(float64[:,:](complex128[:,:,:], complex128[:], float64[:,:], float64[:], float64[:,:], complex128[:]))
def vectorized_NumbaJit_nopythonTrue_DeclareInput(csm, e, h, r0, rm, kj):
    """ In addition to 'vectorized_NumbaJit_nopythonTrue' the in-/output of the
    method are declared in the decorator, which normally leads to speed 
    improvements (even though they're very little in this  particular  case).
    """
    nFreqs = csm.shape[0]
    nGridPoints = r0.shape[0]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.float64)
    for cntFreqs in xrange(nFreqs):
        for cntGrid in xrange(nGridPoints):
            steeringVector = rm[cntGrid, :] / r0[cntGrid] * np.exp(-1j * kj[cntFreqs].imag * (rm[cntGrid, :] - r0[cntGrid])) 
            beamformOutput[cntFreqs, cntGrid] = np.vdot(steeringVector, np.dot(steeringVector, csm[cntFreqs, :, :])).real
    return beamformOutput

@njit(float64[:,:](complex128[:,:,:], float64[:], float64[:,:], complex128[:]), parallel=True)
def vectorizedOptimized_NumbaJit_Parallel(csm, r0, rm, kj):
    """ The parallel=True flag turns on an automized parallezation process.
    When one wants to manually parallelize a certain loop one can do so by
    using prange instead of xrange/range. BUT in this method the prange
    produced errors. Maybe thats because the numpy package performs 
    parallelization itself, which is then in conflict with prange.
    """
    nFreqs = csm.shape[0]
    nGridPoints = r0.shape[0]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.float64)
    for cntFreqs in xrange(nFreqs):
        kjj = kj[cntFreqs].imag
        for cntGrid in xrange(nGridPoints):  # error when trying with prange
            temp3 = (kjj * (rm[cntGrid, :] - r0[cntGrid]))
            steeringVector = rm[cntGrid, :] / r0[cntGrid] * (np.cos(temp3) - 1j * np.sin(temp3))
            beamformOutput[cntFreqs, cntGrid] = (np.vdot(steeringVector, np.dot(steeringVector, csm[cntFreqs, :, :]))).real
    return beamformOutput

#%% Formulate the Beamformer as LOOPS
def loops_exactCopyOfCPP(csm, e, h, r0, rm, kj):
    """ A python copy of the current benchmark function, created with scipy.weave
    """
    nFreqs = csm.shape[0]
    nGridPoints = r0.shape[0]
    nMics = csm.shape[1]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.complex128)
    steerVec = np.zeros((nMics), np.complex128)
    
    for cntFreqs in xrange(nFreqs):
        kjj = kj[cntFreqs].imag
        for cntGrid in xrange(nGridPoints):
            rs = 0
            r01 = r0[cntGrid]
            
            # Calculating of Steering-Vectors
            for cntMics in xrange(nMics):
                rm1 = rm[cntGrid, cntMics]
                rs += 1.0 / (rm1**2)
                temp3 = np.float32(kjj * (rm1 - r01))
                steerVec[cntMics] = (np.cos(temp3) - 1j * np.sin(temp3)) * rm1
            rs = r01 ** 2
            
            # Calculating of the matrix - vector - multiplication
            temp1 = 0.0
            for cntMics in xrange(nMics):
                temp2 = 0.0
                for cntMics2 in xrange(cntMics):
                    temp2 += csm[cntFreqs, cntMics2, cntMics] * steerVec[cntMics2]
                temp1 += 2 * (temp2 * steerVec[cntMics].conjugate()).real
                temp1 += (csm[cntFreqs, cntMics, cntMics] * np.conjugate(steerVec[cntMics]) * steerVec[cntMics]).real
            beamformOutput[cntFreqs, cntGrid] = temp1 / rs
    return beamformOutput

@njit(float64[:,:](complex128[:,:,:], complex128[:], float64[:,:], float64[:], float64[:,:], complex128[:]))
def loops_NumbaJit_nopythonTrue_exactCopyOfCPP(csm, e, h, r0, rm, kj):
    """ See 'vectorized_NumbaJit_nopythonTrue_DeclareInput' for explenation of 
    the numba decorator.
    """
    nFreqs = csm.shape[0]
    nGridPoints = r0.shape[0]
    nMics = csm.shape[1]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.float64)
    steerVec = np.zeros((nMics), np.complex128)
    
    for cntFreqs in xrange(nFreqs):
        kjj = kj[cntFreqs].imag
        for cntGrid in xrange(nGridPoints):
            rs = 0
            r01 = r0[cntGrid]
            for cntMics in xrange(nMics):
                rm1 = rm[cntGrid, cntMics]
                rs += 1.0 / (rm1**2)
                temp3 = np.float32(kjj * (rm1 - r01))
                steerVec[cntMics] = (np.cos(temp3) - 1j * np.sin(temp3)) * rm1
            rs = r01 ** 2
            
            temp1 = 0.0
            for cntMics in xrange(nMics):
                temp2 = 0.0
                for cntMics2 in xrange(cntMics):
                    temp2 += csm[cntFreqs, cntMics2, cntMics] * steerVec[cntMics2]
                temp1 += 2 * (temp2 * steerVec[cntMics].conjugate()).real
                temp1 += (csm[cntFreqs, cntMics, cntMics] * np.conjugate(steerVec[cntMics]) * steerVec[cntMics]).real
            beamformOutput[cntFreqs, cntGrid] = temp1 / rs
    return beamformOutput


@njit(float64[:,:](complex128[:,:,:], complex128[:], float64[:,:], float64[:], float64[:,:], complex128[:]), parallel=True)
def loops_NumbaJit_parallel_FirstWritingOfSteer(csm, e, h, r0, rm, kj):
    """ This method implements the parallelized loop (prange) over the 
    Gridpoints, which is a direct implementation of the currently used 
    c++ method created with scipy.wave.
    
    Very strange: Just like with Cython, this implementation (prange over Gridpoints)
    produces wrong results. If one doesn't parallelize -> everything is good 
    (just like with Cython). Maybe Cython and Numba.jit use the same interpreter 
    to generate OpenMP-parallelizable code.
    
    BUT: If one uncomments the 'steerVec' declaration in the prange-loop over the
    gridpoints an error occurs. After commenting the line again and executing
    the script once more, THE BEAMFORMER-RESULTS ARE CORRECT (for repeated tries). 
    Funny enough the method is now twice as slow in comparison to the 
    'wrong version' (before invoking the error).
    
    A workaround is given by 'loops_NumbaJit_parallel', which is much slower, 
    because the sterring vector is calculated redundantly.
    """
    nFreqs = csm.shape[0]
    nGridPoints = r0.shape[0]
    nMics = csm.shape[1]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.float64)
    steerVec = np.zeros((nMics), np.complex128)
    
    for cntFreqs in xrange(nFreqs):
        kjj = kj[cntFreqs].imag
        for cntGrid in prange(nGridPoints):
#            steerVec = np.zeros((nMics), np.complex128)  # This is the line that has to be uncommented (see this methods documentation comment)
            rs = 0
            r01 = r0[cntGrid]
            
            for cntMics in xrange(nMics):
                rm1 = rm[cntGrid, cntMics]
                rs += 1.0 / (rm1**2)
                temp3 = np.float32(kjj * (rm1 - r01))
                steerVec[cntMics] = (np.cos(temp3) - 1j * np.sin(temp3)) * rm1
            rs = r01 ** 2
            
            temp1 = 0.0
            for cntMics in xrange(nMics):
                temp2 = 0.0
                for cntMics2 in xrange(cntMics):
                    temp2 = temp2 + csm[cntFreqs, cntMics2, cntMics] * steerVec[cntMics2]
                temp1 = temp1 + 2 * (temp2 * steerVec[cntMics].conjugate()).real
                temp1 = temp1 + (csm[cntFreqs, cntMics, cntMics] * np.conjugate(steerVec[cntMics]) * steerVec[cntMics]).real
            
            beamformOutput[cntFreqs, cntGrid] = temp1 / rs
    return beamformOutput

@njit(float64[:,:](complex128[:,:,:], complex128[:], float64[:,:], float64[:], float64[:,:], complex128[:]), parallel=True)
def loops_NumbaJit_parallel(csm, e, h, r0, rm, kj):
    """ Workaround for the prange error in jit. See documentation comment of
    'loops_NumbaJit_parallel_FirstWritingOfSteer'.
    For infos on the numba decorator see 'vectorizedOptimized_NumbaJit_Parallel'
    """
    nFreqs = csm.shape[0]
    nGridPoints = r0.shape[0]
    nMics = csm.shape[1]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.float64)
    
    for cntFreqs in xrange(nFreqs):
        kjj = kj[cntFreqs].imag
        for cntGrid in prange(nGridPoints):
            r01 = r0[cntGrid]
            rs = r01 ** 2
            
            temp1 = 0.0
            for cntMics in xrange(nMics): 
                temp2 = 0.0
                rm1 = rm[cntGrid, cntMics]
                temp3 = np.float32(kjj * (rm1 - r01))
                steerVec = (np.cos(temp3) - 1j * np.sin(temp3)) * rm1
                
                for cntMics2 in xrange(cntMics):
                    rm1 = rm[cntGrid, cntMics2]
                    temp3 = np.float32(kjj * (rm1 - r01))
                    steerVec1 = (np.cos(temp3) - 1j * np.sin(temp3)) * rm1  # Steering vec is calculated redundantly--> very slow
                    temp2 += csm[cntFreqs, cntMics2, cntMics] * steerVec1
                temp1 += 2 * (temp2 * steerVec.conjugate()).real
                temp1 += (csm[cntFreqs, cntMics, cntMics] * steerVec.conjugate() * steerVec).real
            beamformOutput[cntFreqs, cntGrid] = temp1 / rs
    return beamformOutput

#%% Multithreading

#Due to pythons global interpreter lock (GIL) only one thread can run at a time.
#This means, that if one wants to make use of multiple cores, one has to release
#the GIL for concurrently running threads. Numbas jit can release the gil, if 
#all datatypes are primitive (nopython=True) via nogil=True.
#This doesn't have to be done with all the numba.guvectorized stuff, as the 
#multithreading happens automatically.
#I tested

# VERTIG MACHEN





nThreadsGlobal = 2  # einmal mit 2 und einmal mit 4 probieren.. vermutung: die saceh
# die numpy parallelsieirt (ohne jit) arbeitet eh auf beiden cores-> mehr threads bringt dann nichts meh

def vectorized_multiThreading(csm, e, h, r0, rm, kj):
    """ Prepares the Multithreading of 'vectorized_multiThreading_CoreFunction'.
    This method does not free the GIL. As descripted above (beginning of 
    Multithreading section) it therefore shouldn't run concurrently (on multiple 
    cores). BUT as numpys mkl package organizes concurrency itself (see 'vectorized'),
    this
    """
    nThreads = nThreadsGlobal
    dataSizePerThread = nGridPoints / nThreads
    startingIndexPerThread = [cnt * dataSizePerThread for cnt in range(nThreads + 1)]
    startingIndexPerThread[-1] = nGridPoints
    threads = [threading.Thread(target=vectorized_multiThreading_CoreFunction, args=(csm, e, h, r0, rm, kj, startingIndexPerThread[cnt], startingIndexPerThread[cnt+1])) for cnt in range(nThreads)]
    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()
    return h
def vectorized_multiThreading_CoreFunction(csm, e, h, r0, rm, kj, startPoint, endPoint):
    nFreqs = csm.shape[0]
    for cntFreqs in xrange(nFreqs):
        for cntGrid in xrange(startPoint, endPoint):
            steeringVector = rm[cntGrid, :] / r0[cntGrid] * np.exp(-1j * kj[cntFreqs].imag * (rm[cntGrid, :] - r0[cntGrid]))
            h[cntFreqs, cntGrid] = (np.vdot(steeringVector, np.dot(steeringVector, csm[cntFreqs, :, :]))).real


def vectorized_NumbaJit_multiThreading(csm, e, h, r0, rm, kj):
    """ Prepares the Multithreading of 'vectorized_NumbaJit_multiThreading_CoreFunction'
    """
    nThreads = nThreadsGlobal
    dataSizePerThread = nGridPoints / nThreads
    startingIndexPerThread = [cnt * dataSizePerThread for cnt in range(nThreads + 1)]
    startingIndexPerThread[-1] = nGridPoints
    threads = [threading.Thread(target=vectorized_NumbaJit_multiThreading_CoreFunction, args=(csm, e, h, r0, rm, kj, startingIndexPerThread[cnt], startingIndexPerThread[cnt+1])) for cnt in range(nThreads)]
    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()
    return h
@njit(void(complex128[:,:,:], complex128[:], float64[:,:], float64[:], float64[:,:], complex128[:], uint64, uint64), nogil=True)
def vectorized_NumbaJit_multiThreading_CoreFunction(csm, e, h, r0, rm, kj, startPoint, endPoint):
    nFreqs = csm.shape[0]
    for cntFreqs in xrange(nFreqs):
        for cntGrid in xrange(startPoint, endPoint):
            steeringVector = rm[cntGrid, :] / r0[cntGrid] * np.exp(-1j * kj[cntFreqs].imag * (rm[cntGrid, :] - r0[cntGrid]))
            h[cntFreqs, cntGrid] = (np.vdot(steeringVector, np.dot(steeringVector, csm[cntFreqs, :, :]))).real


def loops_NumbaJit_multiThreading(csm, e, h, r0, rm, kj):
    """ Prepares the Multithreading of 'loops_NumbaJit_multiThreading_CoreFunction'.
    Here the cores are used as they should which means: 
    You spawn 2 threads -> cpu uses 2 cores,
    You spawn 3 threads -> cpu uses 3 cores...
    """
    nThreads = nThreadsGlobal
    dataSizePerThread = nGridPoints / nThreads
    startingIndexPerThread = [cnt * dataSizePerThread for cnt in range(nThreads + 1)]
    startingIndexPerThread[-1] = nGridPoints
    threads = [threading.Thread(target=loops_NumbaJit_multiThreading_CoreFunction, args=(csm, e, h, r0, rm, kj, startingIndexPerThread[cnt], startingIndexPerThread[cnt+1])) for cnt in range(nThreads)]
    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()
    return h
@njit(void(complex128[:,:,:], complex128[:], float64[:,:], float64[:], float64[:,:], complex128[:], uint64, uint64), nogil=True)
def loops_NumbaJit_multiThreading_CoreFunction(csm, e, h, r0, rm, kj, startPoint, endPoint):
    nFreqs = csm.shape[0]
    nMics = csm.shape[1]
    steerVec = np.zeros((nMics), np.complex128)
    
    for cntFreqs in xrange(nFreqs):
        kjj = kj[cntFreqs].imag
        for cntGrid in xrange(startPoint, endPoint):
            rs = 0
            r01 = r0[cntGrid]
            for cntMics in xrange(nMics):
                rm1 = rm[cntGrid, cntMics]
                rs += 1.0 / (rm1**2)
                temp3 = np.float32(kjj * (rm1 - r01))
                steerVec[cntMics] = (np.cos(temp3) - 1j * np.sin(temp3)) * rm1
            rs = r01 ** 2
            
            temp1 = 0.0
            for cntMics in xrange(nMics):
                temp2 = 0.0
                for cntMics2 in xrange(cntMics):
                    temp2 += csm[cntFreqs, cntMics2, cntMics] * steerVec[cntMics2]
                temp1 += 2 * (temp2 * steerVec[cntMics].conjugate()).real
                temp1 += (csm[cntFreqs, cntMics, cntMics] * np.conjugate(steerVec[cntMics]) * steerVec[cntMics]).real
            h[cntFreqs, cntGrid] = temp1 / rs

#%% NUMBA - GUVECTORIZE

@guvectorize([void(complex128[:,:], float64[:], float64[:,:], complex128[:], float64[:])], '(m,m),(g),(g,m),()->(g)', nopython=True, target='parallel')
def loops_NumbaGuvectorize(csm, r0, rm, kj, h):
    """ Creating a Numpy-Ufunc: Define it for an input which is n-dimensional.
    Then call it with an input which is n+1 dimensional. Python takes care of 
    the parallelization over the available cores itself.
    In this case python parallelizes over the frequencies.
    Numbas guvectorize doesn't return a value but has to overwrite an result 
    vector passed to the method (in this case 'h') as the last input.
    
    Short description of the guvectorize decorator:
        1. Input-Argument: Declaration of output/input datatypes just like
            with jit, but with a obligatary [] around it
        2. '(m,m),(g)...': A symbolic explenation of the input dimensions. In this
            case 'loops_NumbaGuvectorize' is defined for the following input-dim 
            (csm[nMics x nMics], r0[nGridpoints], rm[nGridpoints x nMics], kj (a scalar), h[nGridpoints])
            , where 'h' contains the calculated results (identified by '->').
            When you then give an input which tensorical order is exactly one order 
            higher then the here made definition (e.g. csm[!nFreqs! x nMics x nMics]), 
            numba automatically distributes the new tensor order onto the 
            muliple cores (in our case every core computes the beamformer map 
            for a single frequency independently of the others)
        3. target: one can compute only on one core (target='CPU'), all available 
            cores (target='parallel') or even on graphic cards (target='cuda') (if drivers are installed)
        4. nopython: See jit-decorator, used above
        
    See also man page "http://numba.pydata.org/".
    
    REMARK: Strangly this seemed only to work, if the added order of CSM was its
        first dimension. E.g. csm[nMics x nMics x nFreqs] didn't seem to work.
    """
    nGridPoints = r0.shape[0]
    nMics = csm.shape[0]
    steerVec = np.zeros((nMics), np.complex128)
    
    kjj = kj[0].imag  # If input is scalar, it has to be dereferenced using the 'variable[0]'-syntax
    for cntGrid in xrange(nGridPoints):
        rs = 0.0
        r01 = r0[cntGrid]
        
        for cntMics in xrange(nMics):
            rm1 = rm[cntGrid, cntMics]
            rs += 1.0 / (rm1**2)
            temp3 = np.float32(kjj * (rm1 - r01))
            steerVec[cntMics] = (np.cos(temp3) - 1j * np.sin(temp3)) * rm1
        rs = r01 ** 2
        
        temp1 = 0.0
        for cntMics in xrange(nMics):
            temp2 = 0.0
            for cntMics2 in xrange(cntMics):
                temp2 += csm[cntMics2, cntMics] * steerVec[cntMics2]
            temp1 += 2 * (temp2 * steerVec[cntMics].conjugate()).real
            temp1 += (csm[cntMics, cntMics] * np.conjugate(steerVec[cntMics]) * steerVec[cntMics]).real
        h[cntGrid] = temp1.real / rs

#@njit(float64[:,:](complex128[:,:,:], float64[:], float64[:,:], complex128[:]))  # right now it doesn't seem to be supported for jit-optimized methods to call guvectorized subroutines. Maybe this will be changed in the future
def loops_NumbaGuvectorizeOverGrid(csm, r0, rm, kj):
    """ Similar to 'loops_NumbaGuvectorize', but in this case the UFunc parallelizes
    over the Gridpoints (as it is done in the scipy.weave version). This leads
    to significant speed improvements. 
    Thoughts on the speed improvements: I can't see why the pipelining should 
    work any more effective in comparison to 'loops_NumbaGuvectorize' (where
    the parallelization is done over the frequency-loop), as in both cases the 
    most time is spend in the loop over the gridpoints, so the chain of 
    instructions should essentially be the same.
    BUT in 'loops_NumbaGuvectorize' the slice of every currently calculated 
    frequency of the CSM is loaded into the shared Cache (e.g. with 4 cores a 
    '4 x nMics x nMics'-tensor is loaded into the shared Cache), whereas with 
    'loops_NumbaGuvectorizeOverGrid' only a '1 x nMics x nMics'-tensor is 
    loaded into the shared Cache. This maybe leads to better managing of resources.
    """
    nGridPoints = r0.shape[0]
    nFreqs = csm.shape[0]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.float64)
    for cntFreqs in xrange(nFreqs):
        result = np.zeros(nGridPoints, np.float64)
        loops_NumbaGuvectorizeOverGrid_CoreFunction(csm[cntFreqs, :, :], r0, rm, kj[cntFreqs], result)
        beamformOutput[cntFreqs, :] = result
    return beamformOutput
    
@guvectorize([(complex128[:,:], float64[:], float64[:], complex128[:], float64[:])], 
              '(m,m),(),(m),()->()', nopython=True, target='parallel')
def loops_NumbaGuvectorizeOverGrid_CoreFunction(csm, r0, rm, kj, h):
    """ CoreFunction of 'loops_NumbaGuvectorizeOverGrid', which does the 
    parallelization over the gridpoints.
    """
    nMics = csm.shape[0]
    steerVec = np.zeros((nMics), np.complex128)
    kjj = kj[0].imag
    
    rs = 0.0
    r01 = r0[0]
    for cntMics in xrange(nMics):
        rm1 = rm[cntMics]
        rs += 1.0 / (rm1**2)
        temp3 = np.float32(kjj * (rm1 - r01))
#==============================================================================
        steerVec[cntMics] = (np.cos(temp3) - 1j * np.sin(temp3)) * rm1
#        steerVec[cntMics] = np.exp(-1j * temp3) * rm1  # is analytically the same as the last line
#
#          With exp(), instead of cos + 1j* sin, the function is noticeable slower 
#          AND the relative error is ca 10^-8 (as with those implementations which 
#          don't perform the down cast from double to 32-bit-precision) 
#          -> Maybe the exp() performs implicitly a cast back to double if its 
#          input is imaginary?!
#==============================================================================
    rs = r01 ** 2
    
    temp1 = 0.0
    for cntMics in xrange(nMics):
        temp2 = 0.0
        for cntMics2 in xrange(cntMics):
            temp2 += csm[cntMics2, cntMics] * steerVec[cntMics2]
        temp1 += 2 * (temp2 * steerVec[cntMics].conjugate()).real
        temp1 += (csm[cntMics, cntMics] * np.conjugate(steerVec[cntMics]) * steerVec[cntMics]).real
    h[0] = temp1 / rs


def loops_NumbaGuvectorizeOverGridNoCast(csm, r0, rm, kj):
    nGridPoints = r0.shape[0]
    nFreqs = csm.shape[0]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.float64)
    for cntFreqs in xrange(nFreqs):
        result = np.zeros(nGridPoints, np.float64)
        loops_NumbaGuvectorizeOverGridNoCast_CoreFunction(csm[cntFreqs, :, :], r0, rm, kj[cntFreqs], result)
        beamformOutput[cntFreqs, :] = result
    return beamformOutput
    
@guvectorize([(complex128[:,:], float64[:], float64[:], complex128[:], float64[:])], 
              '(m,m),(),(m),()->()', nopython=True, target='parallel')
def loops_NumbaGuvectorizeOverGridNoCast_CoreFunction(csm, r0, rm, kj, h):
    nMics = csm.shape[0]
    steerVec = np.zeros((nMics), np.complex128)
    kjj = kj[0].imag
    
    rs = 0.0
    r01 = r0[0]
    for cntMics in xrange(nMics):
        rm1 = rm[cntMics]
        rs += 1.0 / (rm1**2)
        temp3 = kjj * (rm1 - r01)
        steerVec[cntMics] = (np.cos(temp3) - 1j * np.sin(temp3)) * rm1
    rs = r01 ** 2
    
    temp1 = 0.0
    for cntMics in xrange(nMics):
        temp2 = 0.0
        for cntMics2 in xrange(cntMics):
            temp2 += csm[cntMics2, cntMics] * steerVec[cntMics2]
        temp1 += 2 * (temp2 * steerVec[cntMics].conjugate()).real
        temp1 += (csm[cntMics, cntMics] * np.conjugate(steerVec[cntMics]) * steerVec[cntMics]).real
    h[0] = temp1 / rs


def loops_NumbaGuvectorizeOverGridAllCalcsIn32Bit(csm, r0, rm, kj):
    nGridPoints = r0.shape[0]
    nFreqs = csm.shape[0]
    beamformOutput = np.zeros((nFreqs, nGridPoints), np.float64)
    for cntFreqs in xrange(nFreqs):
        result = np.zeros(nGridPoints, np.float64)
#        loops_NumbaGuvectorizeOverGridAllCalcsIn32Bit_CoreFunction(csm[cntFreqs, :, :], r0, rm, kj[cntFreqs], result)
        loops_NumbaGuvectorizeOverGridAllCalcsIn32Bit_CoreFunction(np.complex64(csm[cntFreqs, :, :]), np.float32(r0), np.float32(rm), np.complex64(kj[cntFreqs]), np.float32(result))
        beamformOutput[cntFreqs, :] = result
    return beamformOutput
    
@guvectorize([(complex64[:,:], float32[:], float32[:], complex64[:], float32[:])], '(m,m),(),(m),()->()', nopython=True, target='parallel')
#@guvectorize([(complex128[:,:], float64[:], float64[:], complex128[:], float64[:])], '(m,m),(),(m),()->()', nopython=True, target='parallel')
def loops_NumbaGuvectorizeOverGridAllCalcsIn32Bit_CoreFunction(csm, r0, rm, kj, h):
    nMics = csm.shape[0]
    steerVec = np.zeros((nMics), np.complex64)
    kjj = np.float32(kj[0].imag)
    
    r01 = np.float32(r0[0])
    for cntMics in xrange(nMics):
        rm1 = np.float32(rm[cntMics])
        temp3 = np.float32(kjj * (rm[cntMics] - r01))
        steerVec[cntMics] = np.complex64((np.cos(temp3) - 1j * np.sin(temp3)) * rm1)
    rs = r01 * r01

    temp1 = np.float32(0.0)
#    temp1 = np.float64(0.0)    
    for cntMics in xrange(nMics):
        temp2 = np.complex64(0.0 + 0.0j)
#        temp2 = np.complex128(0.0 + 0.0j)
        for cntMics2 in xrange(cntMics):
            temp2 += csm[cntMics2, cntMics] * steerVec[cntMics2]
        temp1 += 2 * (temp2 * steerVec[cntMics].conjugate()).real
        temp1 += (csm[cntMics, cntMics] * np.conjugate(steerVec[cntMics]) * steerVec[cntMics]).real
    h[0] = temp1 / rs
#%%  MAIN
listOfMics = [64] #[64, 100, 250, 500, 700, 1000]
listGridPoints = [5] # [500, 5000, 10000]  # Standard value: 12000  # The number of gridpoints doesn't seeme to have to great of an influence
nTrials = 10
listOfNFreqs = [20]

#==============================================================================
# The benchmark function 'r_beamfull_inverse' and also other implementations of 
# the beamformer create a lot of overhead, which influences the computational 
# effort of the succeding function. This is mostly the case, if concurrent
# calculations are done (multiple cores). So often the first trial of a new 
# function takes some time longer than the other trials.
#==============================================================================

#funcsToTrial = [vectorized, vectorizedOptimized, vectorized_NumbaJitOnly, \
#                vectorized_NumbaJit_nopythonTrue, vectorized_NumbaJit_nopythonTrue_DeclareInput, \
#                vectorizedOptimized_NumbaJit_Parallel, \
#                loops_exactCopyOfCPP, loops_NumbaJit_nopythonTrue_exactCopyOfCPP, \
#                loops_NumbaJit_parallel_FirstWritingOfSteer, loops_NumbaJit_parallel, \
#                vectorized_multiThreading, vectorized_NumbaJit_multiThreading, loops_NumbaJit_multiThreading, \
#                loops_NumbaGuvectorize, loops_NumbaGuvectorizeOverGrid, \
#                r_beamfull_inverse_OhneMP, r_beamfull_inverse]

#funcsToTrial = [vectorized, vectorizedOptimized, beamformerCythonNOTparallel, loops_NumbaJit_parallel_FirstWritingOfSteer, \
#                vectorized_multiThreading, vectorized_NumbaJit_multiThreading, loops_NumbaJit_multiThreading, \
#                loops_NumbaGuvectorize, loops_NumbaGuvectorizeOverGrid, \
#                r_beamfull_inverse_OhneMP, r_beamfull_inverse]

funcsToTrial = [loops_NumbaGuvectorize, loops_NumbaGuvectorizeOverGrid, loops_NumbaGuvectorizeOverGridNoCast, loops_NumbaGuvectorizeOverGridAllCalcsIn32Bit, r_beamfull_inverse]

for nMics in listOfMics:
    for nGridPoints in listGridPoints:
        for nFreqs in listOfNFreqs:
            # Init
            print(10*'-' + 'New Test configuration: nMics=%s, nGridpoints=%s, nFreqs=%s' %(nMics, nGridPoints, nFreqs) + 10*'-')
            print(10*'-' + 'Creation of inputInputs' + 10*'-')
            
            # Inputs for the beamformer methods:
            # At the moment the beamformer-methods are called once per 
            # frequency (CSM is a Matrix, no 3rd-order-tensor)
            # For easier camparability we build the CSM as a 3rd-order-tensor) instead
            csm = np.random.rand(nFreqs, nMics, nMics) + 1j*np.random.rand(nFreqs, nMics, nMics)  # cross spectral matrix
            for cntFreqs in range(nFreqs):
                csm[cntFreqs, :, :] += csm[cntFreqs, :, :].T.conj()  # make CSM hermetical
            e = np.random.rand(nMics) + 1j*np.random.rand(nMics)  # has no usage
            h = np.zeros((nFreqs, nGridPoints))  # results are stored here, if function has no return value
            r0 = np.random.rand(nGridPoints)  # distance between gridpoints and middle of array
            rm = np.random.rand(nGridPoints, nMics)  # distance between gridpoints and all mics in the array
            kj = np.zeros(nFreqs) + 1j*np.random.rand(nFreqs) # complex wavenumber 
            
            nameOfFuncsToTrial = map(lambda x: x.__name__, funcsToTrial)
            nameOfFuncsForError = [funcName for funcName in nameOfFuncsToTrial if funcName != 'r_beamfull_inverse']
            maxRelativeDeviation = np.zeros((len(funcsToTrial), nTrials))
            maxAbsoluteDeviation = np.zeros((len(funcsToTrial), nTrials))
            timeConsumption = [[] for _ in range(len(funcsToTrial))]
            indOfBaselineFnc = nameOfFuncsToTrial.index('r_beamfull_inverse')
            
            print(10*'-' + 'Onetime calculation of "r_beamfull_inverse" for error reference' + 10*'-')
            r_beamfull_inverse(csm, e, h, r0, rm, kj)
            resultReference = h  # For relative/absolute error
            gc.collect()
            
            # Testing
            print(10*'-' + 'Testing of functions' + 10*'-')
            cntFunc = 0
            for func in funcsToTrial:
                print(func.__name__)
                for cntTrials in xrange(nTrials):
                    h = np.zeros((nFreqs, nGridPoints))
                    if func.__name__ == 'r_beamfull_inverse' or func.__name__ == 'r_beamfull_inverse_OhneMP':
                        t0 = tm.time()
                        func(csm, e, h, r0, rm, kj)
                        t1 = tm.time()
                        result = h
#                        gc.collect()
                    elif func.__name__ == 'loops_NumbaGuvectorize':
                        t0 = tm.time()
                        func(csm, r0, rm, kj, h)
                        t1 = tm.time()
                        result = h
                    elif func.__name__ == 'loops_NumbaGuvectorizeOverGrid' or func.__name__ == 'vectorizedOptimized_NumbaJit_Parallel' or func.__name__ == 'loops_NumbaGuvectorizeOverGridNoCast' or func.__name__ == 'loops_NumbaGuvectorizeOverGridAllCalcsIn32Bit':
                        t0 = tm.time()
                        output = func(csm, r0, rm, kj)
                        t1 = tm.time()
                        result = output
                    elif func.__name__ == 'beamformerCython' or func.__name__ == 'beamformerCythonNOTparallel':
                        t0 = tm.time()
                        output = func(csm, r0, rm, kj)
                        t1 = tm.time()
                        result = np.array(output)
                    else:
                        t0 = tm.time()
                        output = func(csm, e, h, r0, rm, kj)
                        t1 = tm.time()
                        result = output
                    timeConsumption[cntFunc].append(t1 - t0)
                    relativeDiffBetweenNewCodeAndRef = (result - resultReference) / (result + resultReference) * 2  # error in relation to the resulting value
                    maxRelativeDeviation[cntFunc, cntTrials] = np.amax(np.amax(abs(relativeDiffBetweenNewCodeAndRef), axis=1), axis=0)  # relative error in inf-norm
                    maxAbsoluteDeviation[cntFunc, cntTrials] = np.amax(np.amax(abs(result - resultReference), axis=1), axis=0)  # absolute error in inf-norm
                cntFunc += 1
            factorTimeConsump = [np.mean(timeConsumption[cnt]) for cnt in range(0, len(funcsToTrial))] \
                                / np.mean(timeConsumption[indOfBaselineFnc])
            
            # Save the current test-config as .sav
            helpString = 'The order of the variables is: \n nameOfFuncsToTrial \n maxRelativeDeviation'\
                '\n timeConsumption [nFuncs, nTrials] \n nMics \n nGridPoints \n nFreqs '\
                '\n Factor of time consumption (in relation to the original .cpp) \n maxAbsoluteDeviation \n nThreadsGlobal'
            saveTupel = (helpString, nameOfFuncsToTrial, maxRelativeDeviation, timeConsumption, 
                         nMics, nGridPoints, nFreqs, factorTimeConsump, maxAbsoluteDeviation, nThreadsGlobal)
            stringParameters = 'OvernightTestcasesBeamformer_nMics%s_nGridPoints%s_nFreqs%s_nTrials%s' %(nMics, nGridPoints, nFreqs, nTrials)
            
            stringSaveName = 'Peter'
#            stringSaveName = 'Sicherung_DurchgelaufeneTests/Beamformer/AllImportantMethods/' + stringParameters
#            stringSaveName = 'Sicherung_DurchgelaufeneTests/Beamformer/EinflussGridpoints/AMDFX6100/' + stringParameters
#            stringSaveName = 'Sicherung_DurchgelaufeneTests/Beamformer/JitPrange/' + stringParameters
#            stringSaveName = 'Sicherung_DurchgelaufeneTests/Beamformer/Multithreading_02Threads/' + stringParameters

            shFncs.savingTimeConsumption(stringSaveName, saveTupel)  # saving as "stringSaveName.sav"
            
            shFncs.plottingOfOvernightTestcasesBeamformer(stringSaveName + '.sav')  # plot of the current test-config

#==============================================================================
#The Following use of the numba decorators could lead to less code (as a function 
#body could be used more often) but is also slower, which is why it wasn't used
#in this comparison.
# signature = complex128[:,:](complex128[:,:,:], float64[:], float64[:,:])
# numbaOptimizedFunction= jit(signature, nopython=True)(plainPythonFunction.py_func)
#==============================================================================



# -*- coding: utf-8 -*-
#pylint: disable-msg=E0611, E1101, C0103, R0901, R0902, R0903, R0904, W0232
#------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
#------------------------------------------------------------------------------
"""Implements testing of frequency beamformers.
"""

import unittest
import acoular as ac
from traits.api import HasTraits, Int, Float, Bool, Range, TraitEnum, Enum

class Test_Instancing(unittest.TestCase):
    """Test that ensures that digest of Acoular classes changes correctly on
    changes of CArray and List attributes.
    """

    def test_instancing(self):
        """ test that all Acoular classes can be instatiated """
        # iterate over all Acoular definitions labels
        for i in dir(ac):
            with self.subTest(i):
                j = getattr(ac,i) # class, function or variable
                if isinstance(j,type): # is this a class ?
                    j() # this is an instance of the class

    def test_set_traits(self):
        """ test that important traits can be set"""
        # iterate over all Acoular definitions labels
        for i in dir(ac):
            j = getattr(ac,i) # class, function or variable
            # HasTraits derived class ?
            if isinstance(j,type) \
                and issubclass(j,HasTraits) \
                and ('digest' in j.class_traits().keys()):
                do = j.class_traits()['digest'].depends_on
                if do:
                    obj = j()
                    for k in do:
                        with self.subTest(i+'.'+k):
                            if k in j.class_trait_names():
                                tr = j.class_traits()[k]
                                # handling different Trait types
                                # todo: use hypothesis based setattr
                                if tr.is_trait_type(Int):
                                    setattr(obj,k,1)
                                elif tr.is_trait_type(Float):
                                    setattr(obj,k,0.1)
                                elif tr.is_trait_type(Bool):
                                    setattr(obj,k,False)
                                elif tr.is_trait_type(Range):
                                    low = tr.handler._low
                                    high = tr.handler._high
                                    setattr(obj,k,(high+low)/2)
                                elif tr.is_trait_type(TraitEnum) or tr.is_trait_type(Enum):
                                    v = tr.handler.values
                                    setattr(obj,k,v[len(v)//2])

if __name__ == '__main__':
    unittest.main()



from pylab import figure, imshow, colorbar, newaxis, diag, array
from os import path
from acoular import __file__ as bpath, MicGeom, RectGrid, SteeringVector,\
 BeamformerBase, L_p, PowerSpectraImport, ImportGrid
     

# set up the parameters
f = 8000
micgeofile = path.join(path.split(bpath)[0],'xml','array_64.xml')

# source positions and rms values of three sources
loc1=(-0.1,-0.1,0.3) 
loc2=(0.15,0,0.3) 
loc3=(0,0.1,0.3)
rms=array([1,0.7,0.5])

# generate test data, in real life this would come from an array measurement
mg = MicGeom( from_file=micgeofile )
rg = RectGrid( x_min=-0.2, x_max=0.2, y_min=-0.2, y_max=0.2, z=0.3, \
increment=0.01 )
st = SteeringVector(grid=rg, mics=mg)

# obtain the transfer function
st2 = SteeringVector(
    grid=ImportGrid(gpos_file=array([loc1,loc2,loc3]).T), 
    mics=mg)
H = st2.transfer(f).T # transfer functions for 8000 Hz
H_h = H.transpose().conjugate() # H hermetian
Q = diag(rms)**2 # matrix containing the source strength 

# create full csm
csm = (H@Q.astype(complex)@H_h)[newaxis] # calculate csm
ps_import = PowerSpectraImport(csm=csm.copy(), frequencies=f)
bb = BeamformerBase( freq_data=ps_import, steer=st, r_diag=False, cached=False )
pm = bb.synthetic( f, 0 )
Lm = L_p( pm )

# show map
figure()
imshow( Lm.T, origin='lower', vmin=Lm.max()-10, extent=rg.extend(), \
interpolation='bicubic')
colorbar()



# -*- coding: utf-8 -*-
#pylint: disable-msg=E0611, E1101, C0103, R0901, R0902, R0903, R0904, W0232
#------------------------------------------------------------------------------
# Copyright (c) Acoular Development Team.
#------------------------------------------------------------------------------
"""
This file contains NUMBA accelerated functions for time-domain beamformers
"""
import numba as nb
import numpy as np


cachedOption = True  # if True: saves the numba func as compiled func in sub directory
fastOption = True # fastmath options 

@nb.njit([(nb.float64[:,::1], nb.int64[:,::1], nb.float64[:,::1], nb.float64[:,::1], nb.float64[:,::1], nb.float64[:,::1])],
                cache=True, parallel=True, fastmath=True)
def _delayandsum4(data, offsets, ifactor2, steeramp, out, autopower):
    """ Performs one time step of delay and sum with output and additional autopower removal
      
    Parameters
    ----------
    data : float64[nSamples, nMics] 
        The time history for all channels.
    offsets : int64[gridSize, nMics] 
        Indices for each grid point and each channel.
    ifactor2: float64[gridSize, nMics] 
        Second interpolation factor, the first one is computed internally.
    steeramp: float64[gridSize, nMics] 
        Amplitude factor from steering vector.        
    
    Returns
    -------
    None : as the inputs out and autopower get overwritten.
    """
    gridsize, numchannels = offsets.shape
    num = out.shape[0]
    ZERO = data.dtype.type(0.)
    ONE = data.dtype.type(1.)    
    for n in nb.prange(num):
        for gi in nb.prange(gridsize):
            out[n,gi] = ZERO
            autopower[n,gi] = ZERO
            for mi in range(numchannels):
                ind = (gi,mi)
                r = (data[offsets[ind]+n,mi] * (1.-ifactor2[ind]) \
                    + data[offsets[ind]+n+1,mi] * ifactor2[ind]) * steeramp[ind]
                out[n,gi] += r
                autopower[n,gi] += r*r

@nb.njit([(nb.float32[:,::1], nb.int32[:,:,::1], nb.float32[:,:,::1], nb.float32[:,:,::1], nb.float32[:,::1], nb.float32[:,::1]),
            (nb.float64[:,::1], nb.int64[:,:,::1], nb.float64[:,:,::1], nb.float64[:,:,::1], nb.float64[:,::1], nb.float64[:,::1])],
                cache=True, parallel=True, fastmath=True)
def _delayandsum5(data, offsets, ifactor2, steeramp, out, autopower):
    """ Performs one time step of delay and sum with output and additional autopower removal
        
    Parameters
    ----------
    data : float64[nSamples, nMics] 
        The time history for all channels.
    offsets : int64[nBlockSamples, gridSize, nMics] 
        Indices for each grid point and each channel.
    ifactor2: float64[nBlockSamples,gridSize, nMics] 
        Second interpolation factor, the first one is computed internally.
    steeramp: float64[nBlockSamples,gridSize, nMics] 
        Amplitude factor from steering vector.        
    
    Returns
    -------
    None : as the inputs out and autopower get overwritten.
    """
    num, gridsize, numchannels = offsets.shape
    num = out.shape[0]
    #ZERO = data.dtype.type(0.)
    ONE = data.dtype.type(1.)
    for n in nb.prange(num):
        for gi in nb.prange(gridsize):
            out[n,gi] = 0
            autopower[n,gi] = 0
            for mi in range(numchannels):
                ind = offsets[n,gi,mi]+n
                r = (data[ind,mi] * (ONE - ifactor2[n,gi,mi]) \
                    + data[ind+1,mi] * ifactor2[n,gi,mi]) * steeramp[n,gi,mi]
                out[n,gi] += r
                autopower[n,gi] += r*r 

@nb.njit([(nb.float32[:,:,:], nb.float32[:,:], nb.float32[:,:,:]),
            (nb.float64[:,:,:], nb.float64[:,:], nb.float64[:,:,:])],
                cache=True, parallel=True, fastmath=True)
def _steer_I(rm, r0, amp):
    num, gridsize, numchannels = rm.shape
    amp[0,0,0] = 1.0/numchannels# to get the same type for rm2 as for rm
    Nr = amp[0,0,0]
    for n in nb.prange(num): 
        for gi in nb.prange(gridsize):
            for mi in nb.prange(numchannels):
                amp[n,gi,mi] = Nr

@nb.njit([(nb.float32[:,:,:], nb.float32[:,:], nb.float32[:,:,:]),
            (nb.float64[:,:,:], nb.float64[:,:], nb.float64[:,:,:])],
                cache=True, parallel=True, fastmath=True)
def _steer_II(rm, r0, amp):
    num, gridsize, numchannels = rm.shape
    amp[0,0,0] = 1.0/numchannels# to get the same type for rm2 as for rm
    Nr = amp[0,0,0]    
    for n in nb.prange(num): 
        for gi in nb.prange(gridsize):
            rm2 = np.divide(Nr,r0[n,gi])
            for mi in nb.prange(numchannels):
                amp[n,gi,mi] = rm[n,gi,mi]*rm2

@nb.njit([(nb.float32[:,:,:], nb.float32[:,:], nb.float32[:,:,:]),
            (nb.float64[:,:,:], nb.float64[:,:], nb.float64[:,:,:])],
                cache=True, parallel=True, fastmath=True)
def _steer_III(rm, r0, amp):
    num, gridsize, numchannels = rm.shape
    rm20 = rm[0,0,0]-rm[0,0,0] # to get the same type for rm2 as for rm
    rm1 = rm[0,0,0]/rm[0,0,0]
    for n in nb.prange(num): 
        for gi in nb.prange(gridsize):
            rm2 = rm20
            for mi in nb.prange(numchannels):
                rm2 += np.divide(rm1,np.square(rm[n,gi,mi]))
            rm2 *= r0[n,gi]
            for mi in nb.prange(numchannels):
                amp[n,gi,mi] = np.divide(rm1,rm[n,gi,mi]*rm2)

@nb.njit([(nb.float32[:,:,:], nb.float32[:,:], nb.float32[:,:,:]),
            (nb.float64[:,:,:], nb.float64[:,:], nb.float64[:,:,:])],
                cache=True, parallel=True, fastmath=True)
def _steer_IV(rm, r0, amp):
    num, gridsize, numchannels = rm.shape
    amp[0,0,0] = np.sqrt(1.0/numchannels)# to get the same type for rm2 as for rm
    Nr = amp[0,0,0]
    rm1 = rm[0,0,0]/rm[0,0,0]
    rm20 = rm[0,0,0]-rm[0,0,0] # to get the same type for rm2 as for rm
    for n in nb.prange(num): 
        for gi in nb.prange(gridsize):
            rm2 = rm20
            for mi in nb.prange(numchannels):
                rm2 += np.divide(rm1,np.square(rm[n,gi,mi]))
            rm2 = np.sqrt(rm2)
            for mi in nb.prange(numchannels):
                amp[n,gi,mi] = np.divide(Nr,rm[n,gi,mi]*rm2)

@nb.njit([(nb.float32[:,:,::1],  nb.float32, nb.float32[:,:,::1], nb.int32[:,:,::1]),
            (nb.float64[:,:,::1], nb.float64, nb.float64[:,:,::1], nb.int64[:,:,::1])],
                cache=True, parallel=True, fastmath=True)
def _delays(rm, c, interp2, index):
    num, gridsize, numchannels = rm.shape
    invc = 1/c
    intt = index.dtype.type
    for n in nb.prange(num):  
        for gi in nb.prange(gridsize):
            for mi in nb.prange(numchannels):
                delays = invc * rm[n,gi,mi]
                index[n,gi,mi] = intt(delays)
                interp2[n,gi,mi] = delays - nb.int64(delays)

@nb.njit([(nb.float32[:,:,:], nb.float32[:,:,:], nb.int32[:,:,:]),
            (nb.float64[:,:,:], nb.float64[:,:,:], nb.int64[:,:,:])],
                cache=True, parallel=True, fastmath=True)
def _modf(delays, interp2, index):
    num, gridsize, numchannels = delays.shape
    for n in nb.prange(num): 
        for gi in nb.prange(gridsize):
            for mi in nb.prange(numchannels):
                index[n,gi,mi] = int(delays[n,gi,mi])
                interp2[n,gi,mi] = delays[n,gi,mi] - index[n,gi,mi]
     


if __name__ == '__main__':
    foo = _delays
    print(foo.parallel_diagnostics(level=4))
